import traceback
from remi import gui


def default_icon(name, view_w=2, view_h=0.6):
    """
    A simple function to make a default svg icon for the widgets
      such icons can be replaced later with a good one
    """
    icon = gui.Svg(width=100,height=30)
    icon.set_viewbox(-view_w/2,-view_h/2,view_w,view_h)
    text = gui.SvgText(0,0,name)
    text.style['font-size'] = "0.2px"
    text.style['text-anchor'] = "middle"
    stroke_width = 0.01
    rect = gui.SvgRectangle(-view_w/2+stroke_width,-view_h/2+stroke_width,view_w-stroke_width*2,view_h-stroke_width*2)
    rect.set_fill("none")
    rect.set_stroke(0.01,'black')
    icon.append([rect, text])
    return icon

try:
    from .toolbox_EPICS import EPICSBooleanButton, EPICSLed, EPICSValueMeterWidget, EPICSPlotPV, EPICSValueGaugeWidget
except Exception:
    class EPICSPlaceholder(gui.Label):
        icon = default_icon("missing EPICS")
        def __init__(self, msg="In order to use EPICS widgets install pyepics \n" + traceback.format_exc()):
            super(EPICSPlaceholder, self).__init__(msg)
            self.css_white_space = 'pre'

try:
    from .toolbox_opencv import OpencvImRead, OpencvCrop, OpencvVideo, OpencvThreshold, OpencvSplit, OpencvCvtColor, OpencvAddWeighted, OpencvBitwiseNot, OpencvBitwiseAnd, OpencvBitwiseOr,\
                            OpencvBilateralFilter, OpencvBlurFilter, OpencvDilateFilter, OpencvErodeFilter, OpencvLaplacianFilter, OpencvCanny, OpencvFindContours, OpencvInRangeGrayscale,\
                            OpencvMatchTemplate
except Exception:
    class OPENCVPlaceholder(gui.Label):
        icon = default_icon("missing OPENCV")
        def __init__(self, msg="In order to use OpenCv widgets install python-opencv \n" + traceback.format_exc()):
            super(OPENCVPlaceholder, self).__init__(msg)
            self.css_white_space = 'pre'

from .toolbox_scheduling import TimerWidget

try:
    from .toolbox_siemens import PLCSiemens, SiemensButton, BitStatusWidget, WordEditWidget, ByteViewWidget
except Exception:
    class SIEMENSPlaceholder(gui.Label):
        icon = default_icon("missing SIEMENS")
        def __init__(self, msg="In order to use Siemens widgets install python-snap7 \n" + traceback.format_exc()):
            super(SIEMENSPlaceholder, self).__init__(msg)
            self.css_white_space = 'pre'



"""
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import remi
import remi.gui as gui
import inspect
import re
import logging
import types
import traceback
import os


class InstancesTree(gui.TreeView):
    def __init__(self, **kwargs):
        super(InstancesTree, self).__init__(**kwargs)

    def append_instance(self, instance, parent):
        item = gui.TreeItem(instance.variable_name)
        if parent == None:
            parent = self
        item.instance = instance
        item.onclick.do(self.on_tree_item_selected, js_stop_propagation=True)
        parent.append(item)
        return item

    def select_instance(self, node, instance):
        if not hasattr(node, 'attributes'):
            return
        if node.identifier != self.identifier:
            if hasattr(node, 'instance'):
                if node.instance.identifier == instance.identifier:
                    node.style['background-color'] = 'lightblue'
                else:
                    node.style['background-color'] = 'white'
                node.attributes['treeopen'] = 'true'
        for item in node.children.values():
            self.select_instance(item, instance)

    @gui.decorate_event
    def on_tree_item_selected(self, emitter):
        self.select_instance(self, emitter.instance)
        return (emitter.instance,)

    def append_instances_from_tree(self, node, parent=None):
        if not hasattr(node, 'attributes'):
            return
        if not (hasattr(node, 'variable_name') and not node.variable_name is None):
            return
        nodeTreeItem = self.append_instance(node, parent)
        for child in node.children.values():
            self.append_instances_from_tree(child, nodeTreeItem)


class InstancesWidget(gui.VBox):
    def __init__(self, **kwargs):
        super(InstancesWidget, self).__init__(**kwargs)
        self.titleLabel = gui.Label('Instances list', width='100%', height=20)
        self.titleLabel.add_class("DialogTitle")
        self.style['align-items'] = 'flex-start'
        
        self.container = gui.VBox(width="100%", height="calc(100% - 20px)")
        self.container.css_align_items = 'flex-start'
        self.container.css_justify_content = 'flex-start'
        self.container.css_overflow = 'auto'

        self.treeView = InstancesTree()
        self.container.append(self.treeView)

        self.append([self.titleLabel, self.container])

        self.titleLabel.style['order'] = '-1'
        self.titleLabel.style['-webkit-order'] = '-1'
        self.container.style['order'] = '0'
        self.container.style['-webkit-order'] = '0'

    def update(self, editorProject, selectedNode):
        self.treeView.empty()
        if 'root' in editorProject.children.keys():
            self.treeView.append_instances_from_tree(
                editorProject.children['root'])
            self.treeView.select_instance(self.treeView, selectedNode)

    def select(self, selectedNode):
        self.treeView.select_instance(self.treeView, selectedNode)


class ToolBar(gui.HBox):
    def __init__(self, **kwargs):
        super(ToolBar, self).__init__(**kwargs)
        self.css_align_items = 'center'
        self.css_justify_content = 'flex-start'
        self.style['background-color'] = 'white'

    def add_command(self, imagePath, callback, title):
        icon = gui.Image(imagePath, height='90%', margin='0px 1px')
        icon.style['outline'] = '1px solid lightgray'
        icon.onclick.do(callback)
        icon.attributes['title'] = title
        self.append(icon)


class ClassEventConnectorEditor(gui.ClassEventConnector):
    """ This class allows to manage the events. Decorating a method with *decorate_event* decorator
        The method gets the __is_event flag. At runtime, the methods that has this flag gets replaced
        by a ClassEventConnector. This class overloads the __call__ method, where the event method is called,
        and after that the listener method is called too.
    """
    editor_listener_callback = None  # this is the event listener setup by the editor that will receive the events first

    def __call__(self, *args, **kwargs):
        # here the event method gets called
        callback_params = self.event_method_bound(*args, **kwargs)

        if not self.editor_listener_callback is None:
            self.editor_listener_callback(
                self.event_source_instance, *callback_params, **self.kwuserdata)

        if not self.callback:
            return callback_params
        if not callback_params:
            callback_params = self.userdata
        else:
            callback_params = callback_params + self.userdata

        # here the listener gets called, passing as parameters the return values of the event method
        # plus the userdata parameters
        return self.callback(self.event_source_instance, *callback_params, **self.kwuserdata)


class SignalConnection(gui.HBox):
    def __init__(self, widget, listenersList, eventConnectionFuncName, eventConnectionFunc, **kwargs):
        super(SignalConnection, self).__init__(**kwargs)

        self.style.update(
            {'overflow': 'visible', 'height': '24px', 'outline': '1px solid lightgray'})
        self.label = gui.Label(eventConnectionFuncName, width='32%')
        self.label.style.update({'float': 'left', 'font-size': '10px',
                                 'overflow': 'hidden', 'outline': '1px solid lightgray'})

        self.label_do = gui.Label(".do ->", style={'white-space':'nowrap'})

        self.dropdownListeners = gui.DropDown(width='32%', height='100%')
        self.dropdownListeners.onchange.do(self.on_listener_selection)
        self.dropdownListeners.attributes['title'] = "The listener who will receive the event"

        self.dropdownMethods = gui.DropDown(width='32%', height='100%')
        self.dropdownMethods.onchange.do(self.on_connection)
        self.dropdownMethods.attributes['title'] = """The listener's method who will receive the event. \
        A custom method is selected by default. You can select another method, but you should check the method parameters."""

        self.eventConnectionFunc = eventConnectionFunc
        self.eventConnectionFuncName = eventConnectionFuncName
        self.refWidget = widget
        self.listenersList = listenersList
        self.dropdownListeners.append(gui.DropDownItem("None"))
        for w in listenersList:
            ddi = gui.DropDownItem(w.variable_name)
            ddi.listenerInstance = w
            self.dropdownListeners.append(ddi)

        if not self.eventConnectionFunc.callback is None:
            try:
                connectedListenerName = ''
                connectedListenerFunction = None
                print(str(type(eventConnectionFunc.callback)))

                if issubclass(type(eventConnectionFunc.callback), gui.ClassEventConnector):
                    connectedListenerName = eventConnectionFunc.callback.event_method_bound.__self__.variable_name
                    connectedListenerFunction = eventConnectionFunc.callback.event_method_bound
                else:
                    connectedListenerName = eventConnectionFunc.callback.__self__.variable_name
                    connectedListenerFunction = eventConnectionFunc.callback

                self.dropdownListeners.select_by_value(connectedListenerName)
                # this to automatically populate the listener methods dropdown

                self.on_listener_selection(
                    self.dropdownListeners, connectedListenerName)
                print("connected function name:" +
                      connectedListenerFunction.__name__)
                self.dropdownMethods.select_by_value(
                    connectedListenerFunction.__name__)
                # force the connection
                #self.on_connection(None, None)
            except Exception:
                print(traceback.format_exc())
                print(dir(eventConnectionFunc.callback))
                self.disconnect()

        self.append([self.label, self.label_do, self.dropdownListeners, self.dropdownMethods])

    def on_listener_selection(self, widget, dropDownValue):
        self.dropdownMethods.empty()
        if self.dropdownListeners.get_value() == 'None':
            self.disconnect()
        else:
            listener = self.dropdownListeners._selected_item.listenerInstance

            l = []
            func_members = inspect.getmembers(listener)  # , inspect.ismethod)
            for (name, value) in func_members:
                # if issubclass(type(value), gui.ClassEventConnector):
                #    value = value.event_method_bound
                if name not in ['__init__', 'main', 'idle', 'construct_ui'] and type(value) == types.MethodType or issubclass(type(value), gui.ClassEventConnector):
                    ddi = gui.DropDownItem(name)
                    ddi.listenerInstance = listener
                    ddi.listenerFunction = value
                    l.append(ddi)

            # creating a none element
            ddi = gui.DropDownItem('None')
            self.dropdownMethods.append(ddi)

            # here I create a custom listener for the specific event and widgets, the user can select this or an existing method
            if listener.attr_editor_newclass:
                custom_listener_name = self.eventConnectionFuncName + \
                    "_" + self.refWidget.variable_name
                setattr(listener, custom_listener_name, types.MethodType(
                    copy_func(fakeListenerFunc), listener))
                getattr(
                    listener, custom_listener_name).__func__.__name__ = custom_listener_name
                ddi = gui.DropDownItem(custom_listener_name)
                ddi.listenerInstance = listener
                ddi.listenerFunction = getattr(listener, custom_listener_name)
                ddi.style['color'] = "green"
                ddi.style['font-weight'] = "bolder"
                ddi.attributes['title'] = "automatically generated method"
                self.dropdownMethods.append(ddi)

            self.dropdownMethods.append(l)

    def disconnect(self):
        # the listener is canceled when the user selects None
        getattr(self.refWidget, self.eventConnectionFuncName).do(None)
        #getattr(self.refWidget, self.eventConnectionFuncName).editor_listener_callback = None

    def on_connection(self, widget, dropDownValue):
        if self.dropdownMethods.get_value() == 'None':
            self.disconnect()
            return

        listener = self.dropdownMethods._selected_item.listenerInstance
        kwargs = {}
        if hasattr(getattr(self.refWidget, self.eventConnectionFuncName).event_method_bound, "_js_code"):
            kwargs["js_stop_propagation"] = (self.eventConnectionFuncName not in ("onmousedown", "onmousemove", "onmouseleave", "onkeydown"))
        getattr(self.refWidget, self.eventConnectionFuncName).do(
            self.dropdownMethods._selected_item.listenerFunction, **kwargs)


def copy_func(f):
    """Based on https://stackoverflow.com/questions/13503079/how-to-create-a-copy-of-a-python-function"""
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    #g = functools.update_wrapper(g, f)
    if hasattr(f, "__kwdefaults__"):
        g.__kwdefaults__ = f.__kwdefaults__
    return g


def fakeListenerFunc(self, *args):
    print('event trap')


class SignalConnectionManager(gui.Container):
    """ This class allows to interconnect event signals """

    def __init__(self, *args, **kwargs):
        super(SignalConnectionManager, self).__init__(*args, **kwargs)
        self.label = gui.Label('Signal connections', width='100%')
        self.label.add_class("DialogTitle")
        self.append(self.label)
        self.listeners_list = []

    def build_widget_list_from_tree(self, node):
        self.listeners_list.append(node)
        for child in node.children.values():
            if hasattr(child, 'attributes') and (hasattr(child, 'variable_name') and not child.variable_name is None):
                self.build_widget_list_from_tree(child)

    def update(self, widget, widget_tree):
        """ for the selected widget are listed the relative signals
            for each signal there is a dropdown containing all the widgets
            the user will select the widget that have to listen a specific event
        """
        self.listeners_list = []
        self.build_widget_list_from_tree(widget_tree)

        self.label.set_text('Signal connections: ' + widget.variable_name)
        #del self.container
        self.container = gui.VBox(width='100%', height='90%')
        self.container.style['justify-content'] = 'flex-start'
        self.container.style['overflow-y'] = 'scroll'

        # for all the events of this widget
        # isclass instead of ismethod because event methods are replaced with ClassEventConnector
        for (setOnEventListenerFuncname, setOnEventListenerFunc) in inspect.getmembers(widget):
            # if the member is decorated by decorate_set_on_listener and the function is referred to this event
            if issubclass(type(setOnEventListenerFunc), gui.ClassEventConnector):
                self.container.append(SignalConnection(widget,
                                                       self.listeners_list,
                                                       setOnEventListenerFuncname,
                                                       setOnEventListenerFunc,
                                                       width='100%'))

        self.append(self.container, 'container')


class ProjectConfigurationDialog(gui.GenericDialog):
    KEY_PRJ_NAME = 'config_project_name'
    KEY_ADDRESS = 'config_address'
    KEY_PORT = 'config_port'
    KEY_MULTIPLE_INSTANCE = 'config_multiple_instance'
    KEY_ENABLE_CACHE = 'config_enable_file_cache'
    KEY_START_BROWSER = 'config_start_browser'
    KEY_RESOURCEPATH = 'config_resourcepath'

    def __init__(self, title='', message=''):
        super(ProjectConfigurationDialog, self).__init__('Project Configuration',
                                                         'Here are the configuration options of the project.', width=500)
        # standard configuration
        self.configDict = {}

        self.configDict[self.KEY_PRJ_NAME] = 'untitled'
        self.configDict[self.KEY_ADDRESS] = '0.0.0.0'
        self.configDict[self.KEY_PORT] = 8081
        self.configDict[self.KEY_MULTIPLE_INSTANCE] = True
        self.configDict[self.KEY_ENABLE_CACHE] = True
        self.configDict[self.KEY_START_BROWSER] = True
        self.configDict[self.KEY_RESOURCEPATH] = "./res/"

        self.add_field_with_label(
            self.KEY_PRJ_NAME, 'Project Name', gui.TextInput())
        self.add_field_with_label(
            self.KEY_ADDRESS, 'IP address', gui.TextInput())
        self.add_field_with_label(
            self.KEY_PORT, 'Listen port', gui.SpinBox(8082, 1025, 65535))
        self.add_field_with_label(
            self.KEY_MULTIPLE_INSTANCE, 'Use single App instance for multiple users', gui.CheckBox(True))
        self.add_field_with_label(
            self.KEY_ENABLE_CACHE, 'Enable file caching', gui.CheckBox(True))
        self.add_field_with_label(
            self.KEY_START_BROWSER, 'Start browser automatically', gui.CheckBox(True))
        self.add_field_with_label(
            self.KEY_RESOURCEPATH, 'Additional resource path', gui.TextInput())
        self.from_dict_to_fields(self.configDict)

    def from_dict_to_fields(self, dictionary):
        for key in self.inputs.keys():
            if key in dictionary.keys():
                self.get_field(key).set_value(str(dictionary[key]))

    def from_fields_to_dict(self):
        self.configDict[self.KEY_PRJ_NAME] = self.get_field(
            self.KEY_PRJ_NAME).get_value()
        self.configDict[self.KEY_ADDRESS] = self.get_field(
            self.KEY_ADDRESS).get_value()
        self.configDict[self.KEY_PORT] = int(
            self.get_field(self.KEY_PORT).get_value())
        self.configDict[self.KEY_MULTIPLE_INSTANCE] = self.get_field(
            self.KEY_MULTIPLE_INSTANCE).get_value()
        self.configDict[self.KEY_ENABLE_CACHE] = self.get_field(
            self.KEY_ENABLE_CACHE).get_value()
        self.configDict[self.KEY_START_BROWSER] = self.get_field(
            self.KEY_START_BROWSER).get_value()
        self.configDict[self.KEY_RESOURCEPATH] = self.get_field(
            self.KEY_RESOURCEPATH).get_value()

    @gui.decorate_event
    def confirm_dialog(self, emitter):
        """event called pressing on OK button.
        """
        # here the user input is transferred to the dict, ready to use
        self.from_fields_to_dict()
        return super(ProjectConfigurationDialog, self).confirm_dialog(self)

    def show(self, baseAppInstance):
        """Allows to show the widget as root window"""
        self.from_dict_to_fields(self.configDict)
        super(ProjectConfigurationDialog, self).show(baseAppInstance)


class EditorFileSelectionDialog(gui.FileSelectionDialog):
    def __init__(self, title='File dialog', message='Select files and folders',
                 multiple_selection=True, selection_folder='.', allow_file_selection=True,
                 allow_folder_selection=True, baseAppInstance=None):
        super(EditorFileSelectionDialog, self).__init__(title,
                                                        message, multiple_selection, selection_folder,
                                                        allow_file_selection, allow_folder_selection)

        self.baseAppInstance = baseAppInstance

    def show(self, *args):
        super(EditorFileSelectionDialog, self).show(self.baseAppInstance)


class EditorFileSaveDialog(gui.FileSelectionDialog):
    def __init__(self, title='File dialog', message='Select files and folders',
                 multiple_selection=True, selection_folder='.',
                 allow_file_selection=True, allow_folder_selection=True, baseAppInstance=None):
        super(EditorFileSaveDialog, self).__init__(title, message, multiple_selection, selection_folder,
                                                   allow_file_selection, allow_folder_selection)

        self.baseAppInstance = baseAppInstance

    def show(self, *args):
        super(EditorFileSaveDialog, self).show(self.baseAppInstance)

    def add_fileinput_field(self, defaultname='untitled'):
        self.txtFilename = gui.TextInput()
        self.txtFilename.onkeydown.do(self.on_enter_key_pressed)
        self.txtFilename.set_text(defaultname)

        self.add_field_with_label("filename", "Filename", self.txtFilename)

    def get_fileinput_value(self):
        return self.get_field('filename').get_value()

    def on_enter_key_pressed(self, widget, value, keycode):
        if keycode == "13":
            self.confirm_value(None)

    @gui.decorate_event
    def confirm_value(self, widget):
        """event called pressing on OK button.
           propagates the string content of the input field
        """
        self.hide()
        params = (self.fileFolderNavigator.pathEditor.get_text(),)
        return params


def default_icon(name, view_w=1, view_h=0.6):
    """
    A simple function to make a default svg icon for the widgets
      such icons can be replaced later with a good one 
    """
    icon = gui.Svg(width=50,height=30)
    icon.set_viewbox(-view_w/2,-view_h/2,view_w,view_h)
    text = gui.SvgText(0,0,name)
    text.attributes['textLength'] = "100%"
    text.attributes['lengthAdjust'] = "spacingAndGlyphs"
    text.style['font-size'] = "0.2px"
    text.style['text-anchor'] = "middle"
    stroke_width = 0.01
    rect = gui.SvgRectangle(-view_w/2+stroke_width,-view_h/2+stroke_width,view_w-stroke_width*2,view_h-stroke_width*2)
    rect.set_fill("none")
    rect.set_stroke(0.01,'black')
    icon.append([rect, text])
    return icon


class WidgetHelper(gui.HBox):
    """ Allocates the Widget to which it refers,
        interfacing to the user in order to obtain the necessary attribute values
        obtains the constructor parameters, asks for them in a dialog
        puts the values in an attribute called constructor
    """

    def __init__(self, appInstance, widgetClass, **kwargs_to_widget):
        self.kwargs_to_widget = kwargs_to_widget
        self.appInstance = appInstance
        self.widgetClass = widgetClass
        super(WidgetHelper, self).__init__()
        self.style.update({'background-color': 'rgb(250,250,250)', 'width': "auto", 'margin':"2px", 
                           "height": "60px", "justify-content": "center", "align-items": "center",
                           'font-size': '12px'})
        if hasattr(widgetClass, "icon"):
            if type(widgetClass.icon) == gui.Svg:
                self.icon = widgetClass.icon
            elif widgetClass.icon == None:
                self.icon = default_icon(self.widgetClass.__name__)
            else:
                icon_file = widgetClass.icon
                self.icon = gui.Image(icon_file, width='auto', margin='2px')
        else:
            icon_file = '/editor_resources:widget_%s.png' % self.widgetClass.__name__
            if os.path.exists(self.appInstance._get_static_file(icon_file)): 
                self.icon = gui.Image(icon_file, width='auto', margin='2px')
            else:
                self.icon = default_icon(self.widgetClass.__name__)

        self.icon.style['max-width'] = '100%'
        self.icon.style['image-rendering'] = 'auto'
        self.icon.attributes['draggable'] = 'false'
        self.icon.attributes['ondragstart'] = "event.preventDefault();"
        self.append(self.icon, 'icon')
        self.append(gui.Label(self.widgetClass.__name__), 'label')
        self.children['label'].style.update({'margin-left':'2px', 'margin-right':'3px'})

        self.attributes.update({'draggable': 'true',
                                'ondragstart': "this.style.cursor='move'; event.dataTransfer.dropEffect = 'move';   event.dataTransfer.setData('application/json', JSON.stringify(['add',event.target.id,(event.clientX),(event.clientY)]));",
                                'ondragover': "event.preventDefault();",
                                'ondrop': "event.preventDefault();return false;"})

        # this dictionary will contain optional style attributes that have to be added to the widget once created
        self.optional_style_dict = {}

        self.onclick.do(self.create_instance)

    def build_widget_name_list_from_tree(self, node):
        if not hasattr(node, 'attributes'):
            return
        if not (hasattr(node, 'variable_name') and not node.variable_name is None):
            return
        self.varname_list.append(node.variable_name)
        for child in node.children.values():
            self.build_widget_name_list_from_tree(child)

    def build_widget_used_keys_list_from_tree(self, node):
        if not hasattr(node, 'attributes'):
            return
        if not (hasattr(node, 'variable_name') and not node.variable_name is None):
            return
        self.used_keys_list.extend(list(node.children.keys()))
        for child in node.children.values():
            self.build_widget_used_keys_list_from_tree(child)

    def on_dropped(self, left, top):
        self.optional_style_dict['left'] = gui.to_pix(left)
        self.optional_style_dict['top'] = gui.to_pix(top)
        self.create_instance(None)

    def create_instance(self, widget):
        """ Here the widget is allocated
        """
        self.varname_list = []
        self.build_widget_name_list_from_tree(self.appInstance.project)
        self.used_keys_list = []
        self.build_widget_used_keys_list_from_tree(self.appInstance.project)
        print("-------------used keys:" + str(self.used_keys_list))
        variableName = ''
        for i in range(0, 1000):  # reasonably no more than 1000 widget instances in a project
            variableName = self.widgetClass.__name__.lower() + str(i)
            if not variableName in self.varname_list and not variableName in self.used_keys_list:
                break

        """
        if re.match('(^[a-zA-Z][a-zA-Z0-9_]*)|(^[_][a-zA-Z0-9_]+)', variableName) == None:
            self.errorDialog = gui.GenericDialog("Error", "Please type a valid variable name.", width=350,height=120)
            self.errorDialog.show(self.appInstance)
            return

        if variableName in self.varname_list:
            self.errorDialog = gui.GenericDialog("Error", "The typed variable name is already used. Please specify a new name.", width=350,height=150)
            self.errorDialog.show(self.appInstance)
            return
        """
        # here we create and decorate the widget
        widget = self.widgetClass(**self.kwargs_to_widget)
        widget.attr_editor_newclass = False
        widget.variable_name = variableName

        for key in self.optional_style_dict:
            widget.style[key] = self.optional_style_dict[key]
        self.optional_style_dict = {}

        self.appInstance.add_widget_to_editor(widget)


class WidgetCollection(gui.Container):
    def __init__(self, appInstance, **kwargs):
        self.appInstance = appInstance
        super(WidgetCollection, self).__init__(**kwargs)
        self.lblTitle = gui.Label("Widgets Toolbox", height=20)
        self.lblTitle.add_class("DialogTitle")
        self.widgetsContainer = gui.HBox(width='100%', height='calc(100% - 20px)')
        self.widgetsContainer.style.update({'overflow-y': 'scroll',
                                            'overflow-x': 'hidden',
                                            'align-items': 'flex-start',
                                            'flex-wrap': 'wrap',
                                            'background-color': 'white'})

        self.append([self.lblTitle, self.widgetsContainer])

        # load all widgets
        self.add_widget_to_collection(gui.HBox, width='250px', height='250px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.VBox, width='250px', height='250px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.Container, width='250px', height='250px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        #self.add_widget_to_collection(gui.GridBox, width='250px', height='250px', style={'top':'20px', 'left':'20px', 'position':'absolute'})
        self.add_widget_to_collection(gui.Button, text="button", width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.TextInput, width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.Label, text="label", width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.ListView, width='100px', height='100px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute', 'border': '1px solid lightgray'})
        self.add_widget_to_collection(gui.ListItem, text='list item')
        self.add_widget_to_collection(gui.DropDown, width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.DropDownItem, text='drop down item')
        self.add_widget_to_collection(gui.Image, width='100px', height='100px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.CheckBoxLabel, text='check box label', width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.CheckBox, width='30px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.SpinBox, width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.Slider, width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.ColorPicker, width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.Date, width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.Link, text='link', url='', width='100px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.Progress, value=0, _max=100, width='130px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.FileFolderNavigator, width=100, height=100, style = {
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        #self.add_widget_to_collection(gui.VideoPlayer, width='100px', height='100px', style={
        #                              'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.TableWidget, width='100px', height='100px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.TabBox, width='200px', height='200px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(gui.FileUploader, width='150px', height='30px', style={
                                      'top': '20px', 'left': '20px', 'position': 'absolute'})
        self.add_widget_to_collection(
            gui.Svg, style={'top': '20px', 'left': '20px', 'width':'150px', 'height':'150px', 'position': 'absolute', 'border':'1px solid gray'})
        self.add_widget_to_collection(
            gui.SvgSubcontainer, x = '10', y = '10', width = '80', height = '80')
        self.add_widget_to_collection(gui.SvgLine, attributes={
                                      'stroke': 'black', 'stroke-width': '1'})
        self.add_widget_to_collection(gui.SvgCircle)
        self.add_widget_to_collection(gui.SvgEllipse)
        self.add_widget_to_collection(gui.SvgRectangle)
        self.add_widget_to_collection(gui.SvgText)
        self.add_widget_to_collection(gui.SvgPath, attributes={
                                      'stroke': 'black', 'stroke-width': '1'})
        self.add_widget_to_collection(gui.SvgImage)
        self.add_widget_to_collection(gui.SvgGroup)

        #self.load_additional_widgets()

    def load_additional_widgets(self):
        try:
            import widgets
        except Exception:
            from . import widgets
        try:
            classes = inspect.getmembers(widgets, inspect.isclass)
            for (classname, classvalue) in classes:
                if issubclass(classvalue, gui.Widget):
                    self.add_widget_to_collection(
                        classvalue, classvalue.__module__)

        except Exception:
            logging.getLogger('remi.editor').error(
                'error loading external widgets', exc_info=True)

    def add_widget_to_collection(self, widgetClass, group='standard_tools', **kwargs_to_widget):
        # create an helper that will be created on click
        # the helper have to search for function that have 'return' annotation 'event_listener_setter'
        if group not in self.widgetsContainer.children.keys():
            self.widgetsContainer.append(EditorAttributesGroup(group), group)
            self.widgetsContainer.children[group].style['width'] = "100%"

        helper = WidgetHelper(
            self.appInstance, widgetClass, **kwargs_to_widget)
        helper.attributes['title'] = widgetClass.__doc__
        #self.widgetsContainer.append( helper )
        self.widgetsContainer.children[group].append(helper)


class EditorAttributesGroup(gui.VBox):
    """ Contains a title and widgets. When the title is clicked, the contained widgets are hidden.
        Its scope is to provide a foldable group
    """

    def __init__(self, title, **kwargs):
        super(EditorAttributesGroup, self).__init__(**kwargs)
        self.add_class('.RaisedFrame')
        #self.style['display'] = 'block'
        self.container = gui.HBox(width="100%", style={
                                  'overflow': 'visible', 'justify-content': 'flex-start', 'align-items': 'flex-start', 'flex-wrap': 'wrap'})
        self.container.css_justify_content = 'flex-start'
        self.opened = True
        self.title = gui.Label(title, width='100%')
        self.title.add_class("Title")
        self.title.style.update({'text-indent': '25px',
                                 'background-image': "url('/editor_resources:minus.png')",
                                 'background-repeat': 'no-repeat',
                                 'background-position': '5px',
                                 'border-top': '3px solid lightgray'})
        self.title.onclick.do(self.openClose)
        super(EditorAttributesGroup, self).append(self.title)
        super(EditorAttributesGroup, self).append(self.container)

    def openClose(self, widget):
        self.opened = not self.opened
        backgroundImage = "url('/editor_resources:minus.png')" if self.opened else "url('/editor_resources:plus.png')"
        self.title.style['background-image'] = backgroundImage
        self.container.css_display = 'flex' if self.opened else 'none'

    def append(self, widget, key=''):
        return self.container.append(widget, key)

    def remove_child(self, widget):
        return self.container.remove_child(widget)


class EditorAttributes(gui.VBox):
    """ Contains EditorAttributeInput each one of which notify a new value with an event
    """
    __ref_widget = None #The reference widget set by set_widget 

    def __init__(self, appInstance, **kwargs):
        super(EditorAttributes, self).__init__(**kwargs)
        self.appInstance = appInstance
        #self.style['overflow-y'] = 'scroll'
        self.style['justify-content'] = 'flex-start'
        self.style['-webkit-justify-content'] = 'flex-start'
        self.titleLabel = gui.Label('Attributes editor', width='100%', height=20)
        self.titleLabel.add_class("DialogTitle")
        self.infoLabel = gui.Label('Selected widget: None', height=25)
        self.infoLabel.style['font-weight'] = 'bold'
        
        self.attributes_groups_container = gui.VBox(width="100%", height="calc(100% - 45px)")
        self.attributes_groups_container.css_overflow = 'auto'
        self.attributes_groups_container.css_align_items = "flex-start"
        self.attributes_groups_container.css_justify_content = "flex-start"
        
        self.append([self.titleLabel, self.infoLabel, self.attributes_groups_container])

        self.titleLabel.style['order'] = '-1'
        self.titleLabel.style['-webkit-order'] = '-1'
        self.infoLabel.style['order'] = '0'
        self.infoLabel.style['-webkit-order'] = '0'
        self.attributes_groups_container.style['order'] = '1'
        self.attributes_groups_container.style['-webkit-order'] = '1'

        self.group_orders = {
            'Generic': '2', 'WidgetSpecific': '3', 'Geometry': '4', 'Background': '5', 'Transformation': '6'}

        self.attributesInputs = list()
        # load editable attributes
        self.attributeGroups = {}

    def update_widget(self):
        for x, y in inspect.getmembers(self.__ref_widget.__class__):
            if type(y) == property:
                if hasattr(y, "fget"):
                    if hasattr(y.fget, "editor_attributes"):
                        group = y.fget.editor_attributes['group']
                        self.attributeGroups[group].container.children[x].set_value(getattr(self.__ref_widget, x))


    def set_widget(self, widget):

        if not self.__ref_widget is None:
            if self.__ref_widget.identifier == widget.identifier:
                self.update_widget()
                return
        self.__ref_widget = widget

        print("EditorAttributes set widget")
        self.infoLabel.set_text("Selected widget: %s" % widget.variable_name)

        for w in self.attributeGroups.values():
            self.attributes_groups_container.remove_child(w)

        for w in self.attributesInputs:
            if w.attributeDict['group'] in self.attributeGroups:
                self.attributeGroups[w.attributeDict['group']].attributes_groups_container.remove_child(w)

        index = 100
        default_width = "100%"
        default_height = "22px"
        for x, y in inspect.getmembers(self.__ref_widget.__class__):
            if type(y) == property:
                if hasattr(y, "fget"):
                    if hasattr(y.fget, "editor_attributes"):
                        group = y.fget.editor_attributes['group']

                        attributeEditor = None
                        attributeDict = y.fget.editor_attributes
                        # 'background-repeat':{'type':str, 'description':'The repeat behaviour of an optional background image', ,'additional_data':{'possible_values':'repeat | repeat-x | repeat-y | no-repeat | inherit'}},
                        if attributeDict['type'] in (bool, int, float, gui.ColorPicker.__name__, gui.DropDown.__name__, 'url_editor', 'css_size', 'base64_image', 'file'):
                            if attributeDict['type'] == bool:
                                chk = gui.CheckBox(
                                    'checked', width=default_width, height=default_height)
                                attributeEditor = EditorAttributeInputGeneric(
                                    chk, self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == int:
                                spin = gui.SpinBox(attributeDict['additional_data']['default'], attributeDict['additional_data']['min'], attributeDict[
                                                   'additional_data']['max'], attributeDict['additional_data']['step'], width=default_width, height=default_height)
                                attributeEditor = EditorAttributeInputInt(
                                    spin, self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == float:
                                spin = gui.SpinBox(attributeDict['additional_data']['default'], attributeDict['additional_data']['min'], attributeDict[
                                                   'additional_data']['max'], attributeDict['additional_data']['step'], width=default_width, height=default_height)
                                attributeEditor = EditorAttributeInputFloat(
                                    spin, self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == gui.ColorPicker.__name__:
                                attributeEditor = EditorAttributeInputColor(
                                    self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == gui.DropDown.__name__:
                                drop = gui.DropDown(
                                    width=default_width, height=default_height)
                                for value in attributeDict['additional_data']['possible_values']:
                                    drop.append(gui.DropDownItem(value), value)
                                attributeEditor = EditorAttributeInputGeneric(
                                    drop, self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == 'url_editor':
                                attributeEditor = EditorAttributeInputUrl(
                                    self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == 'base64_image':
                                attributeEditor = EditorAttributeInputBase64Image(
                                    self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == 'css_size':
                                attributeEditor = EditorAttributeInputCssSize(
                                    self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)
                            elif attributeDict['type'] == 'file':
                                attributeEditor = EditorAttributeInputFile(
                                    self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)

                        else:  
                            #None is not visible
                            if attributeDict['type'] is None:
                                continue
                            # default editor is string
                            txt = gui.TextInput(
                                width=default_width, height=default_height)
                            attributeEditor = EditorAttributeInputGeneric(
                                txt, self.__ref_widget, x, y, y.fget.editor_attributes, self.appInstance)

                        if not group in self.attributeGroups.keys():
                            groupContainer = EditorAttributesGroup(
                                group, width='100%')
                            groupContainer.css_order = self.group_orders.get(
                                group, str(index))
                            index = index + 1
                            self.attributeGroups[group] = groupContainer

                        if getattr(self.__ref_widget, x) is None:
                            attributeEditor.set_valid(False)
                        else:
                            attributeEditor.set_value(
                                getattr(self.__ref_widget, x))
                        self.attributeGroups[group].append(attributeEditor, key = x)
                        self.attributesInputs.append(attributeEditor)

        for w in self.attributeGroups.values():
            self.attributes_groups_container.append(w)


# widget that allows to edit a specific html and css attributes
#   it has a descriptive label, an edit widget (TextInput, SpinBox..) based on the 'type' and a title
class EditorAttributeInputBase(gui.GridBox):
    """ propertyDef is the property of the class
    """
    targetWidget = None  # wodget
    propertyDef = None  # property
    attributeName = ''
    attributeDict = {}
    appInstance = None
    removeAttribute = None  # widget
    label = None  # widget

    def __init__(self, widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs):
        _style = {'display': 'block',
                  'overflow': 'hidden',
                  'margin': '2px',
                  'outline': '1px solid lightgray',
                  'width': '100%'}
        if 'style' in kwargs.keys():
            kwargs['style'].update(_style)
        else:
            kwargs['style'] = _style

        super(EditorAttributeInputBase, self).__init__(*args, **kwargs)

        self.targetWidget = widget
        self.propertyDef = propertyDef
        self.attributeName = attributeName
        self.attributeDict = attributeDict
        self.appInstance = appInstance
        self.removeAttribute = gui.Image(
            '/editor_resources:delete.png', width='10px')
        self.removeAttribute.attributes['title'] = 'Remove attribute from this widget.'
        self.removeAttribute.onclick.do(self.on_attribute_remove)

        self.label = gui.Label(attributeName, width='100%', height="100%", style={
                               'overflow': 'hidden', 'font-size': '13px', 'margin': '0px'})
        self.label.attributes['title'] = attributeDict['description']

        if not self.propertyDef.fdel is None:
            self.append({'del': self.removeAttribute, 'lbl': self.label})
        else:
            self.append({'lbl': self.label})

        self.set_valid(False)

    def set_valid(self, valid=True):
        self.label.style['opacity'] = '1.0'
        if 'display' in self.removeAttribute.style:
            del self.removeAttribute.style['display']
        if not valid:
            self.label.style['opacity'] = '0.5'
            self.removeAttribute.style['display'] = 'none'

    @gui.decorate_event
    def on_attribute_remove(self, widget):
        self.set_valid(False)
        self.propertyDef.fdel(self.targetWidget)
        return (self.targetWidget, self.attributeName)

    def set_value(self, value):
        self.set_valid(not value is None)
        if not value is None:
            self.inputWidget.set_value(value)

    def on_attribute_changed(self, widget, value):
        self.set_valid()
        setattr(self.targetWidget, self.attributeName, value)


class EditorAttributeInputGeneric(EditorAttributeInputBase):
    inputWidget = None

    def __init__(self, inputWidget, widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs):
        super(EditorAttributeInputGeneric, self).__init__(
            widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs)
        self.inputWidget = inputWidget
        self.inputWidget.onchange.do(self.on_attribute_changed)
        self.inputWidget.attributes['title'] = attributeDict['description']

        '''
        self.set_from_asciiart("""
            |del|lbl                   |input                  |
            """)
        '''
        self.style.update({'grid-template-columns': "6% 46% 48%",
                           'grid-template-rows': "100%", 'grid-template-areas': "'del lbl input'"})
        self.append({'input': self.inputWidget})


class EditorAttributeInputFloat(EditorAttributeInputGeneric):
    def on_attribute_changed(self, emitter, value):
        super(EditorAttributeInputFloat, self).on_attribute_changed(
            self, float(value))


class EditorAttributeInputInt(EditorAttributeInputGeneric):
    def on_attribute_changed(self, emitter, value):
        super(EditorAttributeInputInt, self).on_attribute_changed(
            self, int(float(value)))


class EditorAttributeInputCssSize(EditorAttributeInputBase):
    def __init__(self, widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs):
        super(EditorAttributeInputCssSize, self).__init__(
            widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs)
        self.numInput = gui.SpinBox(
            '0', -999999999, 999999999, 1, width='100%', height='100%')
        self.numInput.onchange.do(self.onchange)
        self.numInput.style['text-align'] = 'right'

        self.dropMeasureUnit = gui.DropDown(width='100%', height='100%')
        self.dropMeasureUnit.append(gui.DropDownItem('px'), 'px')
        self.dropMeasureUnit.append(gui.DropDownItem('%'), '%')
        self.dropMeasureUnit.select_by_key('px')
        self.dropMeasureUnit.onchange.do(self.onchange)
        '''
        self.set_from_asciiart("""
            |del|lbl                   |input           |meas   |
            """)
        '''
        self.style.update({'grid-template-columns': "6% 46% 33% 15%",
                           'grid-template-rows': "100%", 'grid-template-areas': "'del lbl input meas'"})
        self.append({'del': self.removeAttribute, 'lbl': self.label,
                     'input': self.numInput, 'meas': self.dropMeasureUnit})

    def onchange(self, widget, new_value):
        new_size = str(self.numInput.get_value()) + \
            str(self.dropMeasureUnit.get_value())
        self.on_attribute_changed(self, new_size)

    def set_value(self, value):
        """The value have to be in the form '10px' or '10%', so numeric value plus measure unit
        """
        v = 0
        measure_unit = 'px'
        if not value is None:
            try:
                v = int(float(value.replace('px', '')))
            except ValueError:
                try:
                    v = int(float(value.replace('%', '')))
                    measure_unit = '%'
                except ValueError:
                    pass
        self.numInput.set_value(v)
        self.dropMeasureUnit.set_value(measure_unit)
        self.set_valid(not value is None)


class EditorAttributeInputColor(EditorAttributeInputBase):
    def __init__(self, widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs):
        super(EditorAttributeInputColor, self).__init__(
            widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs)
        self.css_height = "60px"
        self.spin_red = gui.SpinBox(0, 0, 255, 1, width="100%", height="100%")
        self.spin_green = gui.SpinBox(
            0, 0, 255, 1, width="100%", height="100%")
        self.spin_blue = gui.SpinBox(0, 0, 255, 1, width="100%", height="100%")
        self.slide_red = gui.Slider(0, 0, 255, 1, width="100%", height="100%", style={
                                    'background-color': 'pink'})
        self.slide_green = gui.Slider(0, 0, 255, 1, width="100%", height="100%", style={
                                      'background-color': 'lightgreen'})
        self.slide_blue = gui.Slider(0, 0, 255, 1, width="100%", height="100%", style={
                                     'background-color': 'lightblue'})
        '''
        self.set_from_asciiart("""
            |del|lbl                   |spin_r  |spin_g  |spin_b  |
            |del|lbl                   |slide_r |slide_g |slide_b |
            """)
        '''
        self.style.update({'grid-template-columns': "6% 46% 16% 16% 16%", 'grid-template-rows': "50% 50%",
                           'grid-template-areas': "'del lbl spin_r spin_g spin_b' 'del lbl slide_r slide_g slide_b'"})
        self.append({'del': self.removeAttribute, 'lbl': self.label, 'spin_r': self.spin_red, 'spin_g': self.spin_green,
                     'spin_b': self.spin_blue, 'slide_r': self.slide_red, 'slide_g': self.slide_green, 'slide_b': self.slide_blue})

        self.slide_red.onchange.do(self.onchange)
        self.slide_green.onchange.do(self.onchange)
        self.slide_blue.onchange.do(self.onchange)

        self.spin_red.onchange.do(self.onchange)
        self.spin_green.onchange.do(self.onchange)
        self.spin_blue.onchange.do(self.onchange)

    def to_str(self):
        return "rgb(%s,%s,%s)" % (self.slide_red.get_value(), self.slide_green.get_value(), self.slide_blue.get_value())

    def from_str(self, value_str):
        components = []
        if value_str is None or '(' not in value_str or ')' not in value_str:
            components = [0, 0, 0]
        else:
            components = value_str[value_str.index(
                '(')+1:value_str.index(')')].split(',')
        if len(components) < 3:
            components = [0, 0, 0]
        self.slide_red.set_value(components[0])
        self.slide_green.set_value(components[1])
        self.slide_blue.set_value(components[2])

        self.spin_red.set_value(self.slide_red.get_value())
        self.spin_green.set_value(self.slide_green.get_value())
        self.spin_blue.set_value(self.slide_blue.get_value())

    @gui.decorate_event
    def onchange(self, widget, new_value):
        if type(widget) == gui.SpinBox:
            self.slide_red.set_value(self.spin_red.get_value())
            self.slide_green.set_value(self.spin_green.get_value())
            self.slide_blue.set_value(self.spin_blue.get_value())
        else:
            self.spin_red.set_value(self.slide_red.get_value())
            self.spin_green.set_value(self.slide_green.get_value())
            self.spin_blue.set_value(self.slide_blue.get_value())
        print("color changed")
        self.set_valid()
        setattr(self.targetWidget, self.attributeName, self.to_str())

    def set_value(self, value):
        self.set_valid(not value is None)
        if not value is None:
            self.from_str(value)


class EditorAttributeInputUrl(EditorAttributeInputBase):
    inputWidget = None

    def __init__(self, widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs):
        super(EditorAttributeInputUrl, self).__init__(widget, attributeName,
                                                      propertyDef, attributeDict, appInstance, *args, **kwargs)
        self.inputWidget = gui.TextInput(width="100%", height="100%")
        self.inputWidget.onchange.do(self.on_attribute_changed)
        self.inputWidget.attributes['title'] = attributeDict['description']

        self.btFileFolderSelection = gui.Widget(width='100%', height='100%')
        self.btFileFolderSelection.style.update({'background-repeat': 'no-repeat',
                                                 'background-image': "url('/res:folder.png')",
                                                 'background-color': 'transparent'})
        self.btFileFolderSelection.onclick.do(
            self.on_file_selection_bt_pressed)
        '''
        self.set_from_asciiart("""
            |del|lbl                   |input                |bt|
            """)
        '''
        self.style.update({'grid-template-columns': "6% 46% 33% 15%",
                           'grid-template-rows': "100%", 'grid-template-areas': "'del lbl input bt'"})
        self.append({'del': self.removeAttribute, 'lbl': self.label,
                     'input': self.inputWidget, 'bt': self.btFileFolderSelection})

    def on_file_selection_bt_pressed(self, widget):
        self.selectionDialog = gui.FileSelectionDialog(
            'Select a file', '', False, './', True, False)
        self.selectionDialog.confirm_value.do(self.file_dialog_confirmed)
        self.selectionDialog.show(self.appInstance)

    def file_dialog_confirmed(self, widget, fileList):
        if len(fileList) > 0:
            self.inputWidget.set_value(
                "url('/editor_resources:" + fileList[0].split('/')[-1].split('\\')[-1] + "')")
            return self.on_attribute_changed(None, self.inputWidget.get_value())

    def set_value(self, value):
        self.set_valid(not value is None)
        if not value is None:
            self.inputWidget.set_value(value)


class EditorAttributeInputBase64Image(EditorAttributeInputUrl):
    def __init__(self, widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs):
        super(EditorAttributeInputBase64Image, self).__init__(
            widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs)

    def file_dialog_confirmed(self, widget, fileList):
        if len(fileList) > 0:
            self.inputWidget.set_value(gui.load_resource(fileList[0]))
            return self.on_attribute_changed(None, self.inputWidget.get_value())


class EditorAttributeInputFile(EditorAttributeInputUrl):
    def __init__(self, widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs):
        super(EditorAttributeInputFile, self).__init__(
            widget, attributeName, propertyDef, attributeDict, appInstance, *args, **kwargs)

    def file_dialog_confirmed(self, widget, fileList):
        if len(fileList) > 0:
            self.inputWidget.set_value(fileList[0].replace("\\", "/"))
            return self.on_attribute_changed(None, self.inputWidget.get_value())



from editor import prototypes
from editor import editor_widgets
try:
    from editor import Editor
except ImportError:
    from editor.editor import Editor

from remi import start

if __name__ == "__main__":
    start(Editor, debug=False, address='0.0.0.0', port=8082, update_interval=0.01, multiple_instance=True)



"""
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import remi.gui as gui
import remi.server
from remi import start, App
import inspect
import sys
import os  # for path handling
try:
    import prototypes
    import editor_widgets
except ImportError:
    from . import prototypes
    from . import editor_widgets

import threading
import traceback

if remi.server.pyLessThan3:
    import imp

    def load_source(filename):
        return imp.load_source('project', filename)
else:
    import importlib.machinery
    import importlib.util

    def load_source(filename):
        loader = importlib.machinery.SourceFileLoader('project', filename)
        spec = importlib.util.spec_from_loader(loader.name, loader)
        _module = importlib.util.module_from_spec(spec)
        loader.exec_module(_module)
        return _module


class DraggableItem(gui.EventSource):
    def __init__(self, app_instance, **kwargs):
        gui.EventSource.__init__(self)
        self.app_instance = app_instance
        self.refWidget = None
        self.parent = None
        self.active = False
        self.origin_x = -1
        self.origin_y = -1
        self.snap_grid_size = 15

    def setup(self, refWidget, newParent):
        # refWidget is the target widget that will be resized
        # newParent is the container
        if self.parent:
            try:
                self.parent.remove_child(self)
            except Exception:
                pass
        if newParent == None:
            return
        self.parent = newParent
        self.refWidget = refWidget

        try:
            self.parent.append(self)
        except Exception:
            pass
        self.update_position()

    def start_drag(self, emitter, x, y):
        self.active = True
        self.app_instance.mainContainer.onmousemove.do(self.on_drag, js_prevent_default=True, js_stop_propagation=False)
        self.app_instance.mainContainer.onmouseup.do(self.stop_drag, js_prevent_default=True, js_stop_propagation=False)
        self.app_instance.mainContainer.onmouseleave.do(self.stop_drag, 0, 0, js_prevent_default=True, js_stop_propagation=False)
        self.origin_x = -1
        self.origin_y = -1

    @gui.decorate_event
    def stop_drag(self, emitter, x, y):
        self.active = False
        self.update_position()
        self.app_instance.mainContainer.onmousemove.do(None, js_prevent_default=False, js_stop_propagation=True)
        self.app_instance.mainContainer.onmouseup.do(None, js_prevent_default=False, js_stop_propagation=True)
        self.app_instance.mainContainer.onmouseleave.do(None, 0, 0, js_prevent_default=False, js_stop_propagation=True)
        return ()

    def on_drag(self, emitter, x, y):
        pass

    def update_position(self):
        pass

    def set_snap_grid_size(self, value):
        self.snap_grid_size = value

    def round_grid(self, value):
        return int(value/self.snap_grid_size)*self.snap_grid_size


class SvgDraggablePoint(gui.SvgRectangle, DraggableItem):
    def __init__(self, app_instance, name_coord_x, name_coord_y, compatibility_iterable, **kwargs):
        self.w = 15
        self.h = 15
        super(SvgDraggablePoint, self).__init__(0, 0, self.w, self.h, **kwargs)
        DraggableItem.__init__(self, app_instance, **kwargs)
        self.attributes['stroke-dasharray'] = "2,2"
        self.name_coord_x = name_coord_x
        self.name_coord_y = name_coord_y
        self.set_stroke(1, 'black')
        self.set_fill('#ffcc00')
        self.compatibility_iterable = compatibility_iterable
        self.onmousedown.do(self.start_drag)

    def setup(self, refWidget, newParent):
        if type(refWidget) in self.compatibility_iterable or refWidget == None:
            DraggableItem.setup(self, refWidget, newParent)

    def on_drag(self, emitter, x, y):
        if self.active:
            if self.origin_x == -1:
                self.origin_x = float(x)
                self.origin_y = float(y)
                self.refWidget_origin_x = float(
                    self.refWidget.attributes[self.name_coord_x])
                self.refWidget_origin_y = float(
                    self.refWidget.attributes[self.name_coord_y])
            else:
                self.refWidget.attributes[self.name_coord_x] = self.round_grid(
                    self.refWidget_origin_x + float(x) - self.origin_x)
                self.refWidget.attributes[self.name_coord_y] = self.round_grid(
                    self.refWidget_origin_y + float(y) - self.origin_y)
                self.update_position()

    def update_position(self):
        if self.refWidget:
            self.set_position(float(self.refWidget.attributes[self.name_coord_x])-self.w/2, float(
                self.refWidget.attributes[self.name_coord_y])-self.h/2)


class SvgDraggableRectangleResizePoint(gui.SvgRectangle, DraggableItem):
    def __init__(self, app_instance, compatibility_iterable, **kwargs):
        self.w = 15
        self.h = 15
        super(SvgDraggableRectangleResizePoint, self).__init__(
            0, 0, self.w, self.h, **kwargs)
        DraggableItem.__init__(self, app_instance, **kwargs)
        self.attributes['stroke-dasharray'] = "2,2"
        self.set_stroke(1, 'black')
        self.set_fill('#ffcc00')
        self.compatibility_iterable = compatibility_iterable
        self.onmousedown.do(self.start_drag)

    def setup(self, refWidget, newParent):
        if type(refWidget) in self.compatibility_iterable or refWidget == None:
            DraggableItem.setup(self, refWidget, newParent)

    def on_drag(self, emitter, x, y):
        if self.active:
            if self.origin_x == -1:
                self.origin_x = float(x)
                self.origin_y = float(y)
                self.refWidget_origin_w = float(
                    self.refWidget.attributes['width'])
                self.refWidget_origin_h = float(
                    self.refWidget.attributes['height'])
            else:
                self.refWidget.attributes['width'] = self.round_grid(
                    self.refWidget_origin_w + float(x) - self.origin_x)
                self.refWidget.attributes['height'] = self.round_grid(
                    self.refWidget_origin_h + float(y) - self.origin_y)
                self.update_position()

    def update_position(self):
        if self.refWidget:
            self.set_position(float(self.refWidget.attributes['x'])+float(self.refWidget.attributes['width'])-self.w/2, float(
                self.refWidget.attributes['y'])+float(self.refWidget.attributes['height'])-self.h/2)


class SvgDraggableCircleResizeRadius(gui.SvgRectangle, DraggableItem):
    def __init__(self, app_instance, compatibility_iterable, **kwargs):
        self.w = 15
        self.h = 15
        super(SvgDraggableCircleResizeRadius, self).__init__(
            0, 0, self.w, self.h, **kwargs)
        DraggableItem.__init__(self, app_instance, **kwargs)
        self.attributes['stroke-dasharray'] = "2,2"
        self.set_stroke(1, 'black')
        self.set_fill('#ffcc00')
        self.compatibility_iterable = compatibility_iterable
        self.onmousedown.do(self.start_drag)

    def setup(self, refWidget, newParent):
        if type(refWidget) in self.compatibility_iterable or refWidget == None:
            DraggableItem.setup(self, refWidget, newParent)

    def on_drag(self, emitter, x, y):
        if self.active:
            if self.origin_x == -1:
                self.origin_x = float(x)
                self.origin_y = float(y)
                self.refWidget_origin_r = float(self.refWidget.attributes['r'])
            else:
                r = self.round_grid(
                    self.refWidget_origin_r + float(x) - self.origin_x)
                self.refWidget.attributes['r'] = str(max(0, r))
                self.update_position()

    def update_position(self):
        if self.refWidget:
            self.set_position(float(self.refWidget.attributes['cx'])+float(
                self.refWidget.attributes['r'])-self.w/2, float(self.refWidget.attributes['cy'])-self.h/2)


class ResizeHelper(gui.Widget, DraggableItem):

    def __init__(self, app_instance, **kwargs):
        super(ResizeHelper, self).__init__(**kwargs)
        DraggableItem.__init__(self, app_instance, **kwargs)
        self.style['float'] = 'none'
        self.style['background-image'] = "url('/editor_resources:resize.png')"
        self.style['background-color'] = "rgba(255,255,255,0.0)"
        self.style['position'] = 'absolute'
        self.style['left'] = '0px'
        self.style['top'] = '0px'
        self.onmousedown.do(self.start_drag)

    def setup(self, refWidget, newParent):
        self.style['display'] = 'none'
        if issubclass(newParent.__class__, gui.TabBox):
            return
        if issubclass(refWidget.__class__, gui.Widget) and 'left' in refWidget.style and 'top' in refWidget.style and \
                'width' in refWidget.style and 'height' in refWidget.style and refWidget.css_position=='absolute':
            DraggableItem.setup(self, refWidget, newParent)
            self.style['display'] = 'block'

    def on_drag(self, emitter, x, y):
        if self.active:
            if self.origin_x == -1:
                self.origin_x = float(x)
                self.origin_y = float(y)
                self.refWidget_origin_w = gui.from_pix(
                    self.refWidget.style['width'])
                self.refWidget_origin_h = gui.from_pix(
                    self.refWidget.style['height'])
            else:
                self.refWidget.style['width'] = gui.to_pix(self.round_grid(
                    self.refWidget_origin_w + float(x) - self.origin_x))
                self.refWidget.style['height'] = gui.to_pix(self.round_grid(
                    self.refWidget_origin_h + float(y) - self.origin_y))
                self.update_position()

    def update_position(self):
        self.style['position'] = 'absolute'
        if self.refWidget:
            if 'left' in self.refWidget.style and 'top' in self.refWidget.style:
                self.style['left'] = gui.to_pix(gui.from_pix(
                    self.refWidget.style['left']) + gui.from_pix(self.refWidget.style['width']))
                self.style['top'] = gui.to_pix(gui.from_pix(
                    self.refWidget.style['top']) + gui.from_pix(self.refWidget.style['height']))


class DragHelper(gui.Widget, DraggableItem):

    def __init__(self, app_instance, **kwargs):
        super(DragHelper, self).__init__(**kwargs)
        DraggableItem.__init__(self, app_instance, **kwargs)
        self.style['float'] = 'none'
        self.style['background-image'] = "url('/editor_resources:drag.png')"
        self.style['background-color'] = "rgba(255,255,255,0.0)"
        self.style['position'] = 'absolute'
        self.style['left'] = '0px'
        self.style['top'] = '0px'
        self.onmousedown.do(self.start_drag)

    def setup(self, refWidget, newParent):
        self.style['display'] = 'none'
        if issubclass(newParent.__class__, gui.TabBox):
            return
        if issubclass(refWidget.__class__, gui.Widget) and 'left' in refWidget.style and 'top' in refWidget.style and refWidget.css_position=='absolute':
            DraggableItem.setup(self, refWidget, newParent)
            self.style['display'] = 'block'

    def on_drag(self, emitter, x, y):
        if self.active:
            if self.origin_x == -1:
                self.origin_x = float(x)
                self.origin_y = float(y)
                self.refWidget_origin_x = gui.from_pix(
                    self.refWidget.style['left'])
                self.refWidget_origin_y = gui.from_pix(
                    self.refWidget.style['top'])
            else:
                self.refWidget.style['left'] = gui.to_pix(self.round_grid(
                    self.refWidget_origin_x + float(x) - self.origin_x))
                self.refWidget.style['top'] = gui.to_pix(self.round_grid(
                    self.refWidget_origin_y + float(y) - self.origin_y))
                self.update_position()

    def update_position(self):
        self.style['position'] = 'absolute'
        if self.refWidget:
            if 'left' in self.refWidget.style and 'top' in self.refWidget.style:
                self.style['left'] = gui.to_pix(gui.from_pix(
                    self.refWidget.style['left'])-gui.from_pix(self.style['width']))
                self.style['top'] = gui.to_pix(gui.from_pix(
                    self.refWidget.style['top'])-gui.from_pix(self.style['width']))


class Project(gui.Container):
    """ The editor project is pure html with specific tag attributes
        This class loads and save the project file, 
        and also compiles a project in python code.
    """

    def __init__(self, **kwargs):
        super(Project, self).__init__(**kwargs)
        self.variable_name = 'App'# + str(id(self)) 
        self.style.update({'position': 'relative',
                           'overflow': 'auto',
                           'background-color': 'rgb(250,248,240)',
                           'background-image': "url('/editor_resources:background.png')"})
        self.attr_editor_newclass = True

    def load(self, ifile, configuration):
        self.ifile = ifile

        _module = load_source(self.ifile)

        configuration.configDict = _module.configuration

        # finding App class
        clsmembers = inspect.getmembers(_module, inspect.isclass)

        app_init_fnc = None
        for (name, value) in clsmembers:
            if issubclass(value, App) and name != 'App':
                app_init_fnc = value

        if app_init_fnc == None:
            return None

        members_list = app_init_fnc.__dict__.values()
        for m in members_list:
            if inspect.isfunction(m) and m.__name__ not in ['__init__', 'main', 'idle', 'construct_ui']:
                import types
                setattr(self, m.__name__, types.MethodType(m, self))
                print(m.__name__)
        root_widget = app_init_fnc.construct_ui(self)
        return root_widget

    def check_pending_listeners(self, widget, widgetVarName, force=False):
        code_nested_listener = ''
        # checking if pending listeners code production can be solved
        for event in self.pending_listener_registration:
            if force or (hasattr(event['eventsource'], 'path_to_this_widget') and hasattr(event['eventlistener'], 'path_to_this_widget')):
                if (force or (widget.variable_name in event['eventsource'].path_to_this_widget and widget.variable_name in event['eventlistener'].path_to_this_widget)) and event['done'] == False:
                    # this means that this is the root node from where the leafs(listener and source) departs, hre can be set the listener
                    event['done'] = True

                    # event source chain
                    sourcename = 'self'
                    source_filtered_path = event['eventsource'].path_to_this_widget[:]
                    listener_filtered_path = event['eventlistener'].path_to_this_widget[:]
                    for v in widget.path_to_this_widget:
                        source_filtered_path.remove(v)
                        listener_filtered_path.remove(v)

                    if len(source_filtered_path) == 0 and event['eventsource'].attr_editor_newclass == False:
                        sourcename = event['eventsource'].variable_name

                    #    listenername = event['eventlistener'].identifier
                    if force or (self.children['root'] == widget and not (widget.attr_editor_newclass == True)):
                        sourcename = self.children['root'].variable_name
                        if self.children['root'].variable_name in source_filtered_path:
                            source_filtered_path.remove(
                                self.children['root'].variable_name)

                    if len(source_filtered_path) > 0:
                        sourcename = (
                            "%s.children['" + "'].children['".join(source_filtered_path) + "']") % sourcename

                    # listener chain
                    listenername = "self"
                    if force or (self.children['root'] == widget and not (widget.attr_editor_newclass == True)):
                        if event['eventlistener'] != self:
                            listenername = self.children['root'].variable_name
                    if len(listener_filtered_path) == 0 and event['eventlistener'].attr_editor_newclass == False:
                        listenername = event['eventlistener'].variable_name
                    if len(listener_filtered_path) > 0:
                        listenername = (
                            "%s.children['" + "'].children['".join(listener_filtered_path) + "']") % listenername

                    code_nested_listener += prototypes.proto_set_listener % {'sourcename': sourcename,
                                                                             'register_function':  event['setoneventfuncname'],
                                                                             'listenername': listenername,
                                                                             'listener_function': event['listenerfuncname']}
                    if not event['eventlistener'].identifier in self.code_declared_classes:
                        self.code_declared_classes[event['eventlistener'].identifier] = ''

                    if (event['eventlistener'].attr_editor_newclass == True):
                        if not event['skip_function_definition']:
                            self.code_declared_classes[event['eventlistener']
                                                   .identifier] += event['listenerClassFunction']
        return code_nested_listener

    # widgetVarName is the name with which the parent calls this instance
    def repr_widget_for_editor(self, widget, first_node=False):
        if first_node:
            self.code_declared_classes = {}
            self.pending_listener_registration = list()
            # a list containing widgets that have been parsed and that are considered valid listeners
            self.known_project_children = [self, ]
            # a list containing dicts {listener, emitter, register_function, listener_function}
            self.pending_signals_to_connect = list()
            self.path_to_this_widget = []
            self.prepare_path_to_this_widget(self.children['root'])
        self.known_project_children.append(widget)

        widget.path_to_this_widget.append(widget.variable_name)

        print(widget.variable_name)

        code_nested = ''  # the code strings to return

        if not hasattr(widget, 'attributes'):
            return ''  # no nested code

        widgetVarName = widget.variable_name
        classname = 'CLASS' + \
            widgetVarName if widget.attr_editor_newclass else widget.__class__.__name__

        code_nested = prototypes.proto_widget_allocation % {
            'varname': widgetVarName, 'classname': classname}

        #code_nested += prototypes.proto_attribute_setup%{'varname': widgetVarName, 'attr_dict': ','.join('"%s":"%s"'%(key,widget.attributes[key]) for key in widget.attributes.keys() if key not in html_helper.htmlInternallyUsedTags)}
        #code_nested += prototypes.proto_style_setup%{'varname': widgetVarName, 'style_dict': ','.join('"%s":"%s"'%(key,widget.style[key]) for key in widget.style.keys())}

        for x, y in inspect.getmembers(widget.__class__):
            if type(y) == property and not getattr(widget, x) is None:
                if hasattr(y.fget, "editor_attributes"): #if this property is visible for the editor
                    _value = getattr(widget, x)
                    if type(_value) == str:
                        _value = '"%s"' % _value
                    code_nested += prototypes.proto_property_setup % {
                        'varname': widgetVarName, 'property': x, 'value': _value}

        # for all the methods of this widget
        for (setOnEventListenerFuncname, setOnEventListenerFunc) in inspect.getmembers(widget):
            # if the member is decorated by decorate_set_on_listener
            if issubclass(type(setOnEventListenerFunc), gui.ClassEventConnector):
                # if there is a callback
                if not setOnEventListenerFunc.callback is None and hasattr(setOnEventListenerFunc.event_method_bound, '_event_info'):
                    listenerFunction = setOnEventListenerFunc.callback
                    if issubclass(type(listenerFunction), gui.ClassEventConnector):
                        listenerFunction = listenerFunction.event_method_bound

                    listenerPrototype = setOnEventListenerFunc.event_method_bound._event_info[
                        'prototype']
                    listener = listenerFunction.__self__

                    # setOnEventListenerFunc._event_info['name'] + "_" + widget.identifier
                    listenerFunctionName = listenerFunction.__name__

                    listenerClassFunction = prototypes.proto_code_function % {'funcname': listenerFunctionName,
                                                                              'parameters': listenerPrototype}

                    # override, if already implemented, we use this code, unless it is a fakeListenerFunction
                    if hasattr(listener, listenerFunctionName) and listenerFunction.__code__ != editor_widgets.fakeListenerFunc.__code__:
                        listenerClassFunction = inspect.getsource(
                            listenerFunction)

                    # if the listener is already in the dictionary, and so it already used by another event
                    # skip
                    skip = False
                    for pending in self.pending_listener_registration:
                        if pending['eventlistener'] == listener:
                            if pending['listenerfuncname'] == listenerFunctionName:
                                skip = True
                                break
                    #if skip:
                    #    continue

                    self.pending_listener_registration.append({'done': False,
                                                               'eventsource': widget,
                                                               'eventlistener': listener,
                                                               'setoneventfuncname': setOnEventListenerFuncname,
                                                               'listenerfuncname': listenerFunctionName,
                                                               'listenerClassFunction': listenerClassFunction,
                                                               'skip_function_definition':skip})

        if widget.attr_editor_newclass:
            widgetVarName = 'self'

        children_code_nested = ''
        for child_key in widget.children.keys():
            child = widget.children[child_key]
            if type(child) == str:
                #children_code_nested += prototypes.proto_layout_append%{'parentname':widgetVarName,'varname':"'%s'"%child}
                continue
            if not issubclass(child.__class__, gui.Widget):
                continue
            if child.variable_name is None:
                continue
            child.path_to_this_widget = widget.path_to_this_widget[:]
            children_code_nested += self.repr_widget_for_editor(child)
            children_code_nested += prototypes.proto_layout_append % {
                'parentname': widgetVarName, 'varname': "%s,'%s'" % (child.variable_name, child.variable_name)}

        children_code_nested += self.check_pending_listeners(
            widget, widgetVarName)

        # and not (classname in self.code_declared_classes.keys()):
        if widget.attr_editor_newclass:
            if not widget.identifier in self.code_declared_classes:
                self.code_declared_classes[widget.identifier] = ''
            self.code_declared_classes[widget.identifier] = prototypes.proto_code_class % {'classname': classname, 'superclassname': widget.__class__.__name__,
                                                                                           'nested_code': children_code_nested} + self.code_declared_classes[widget.identifier]
        else:
            code_nested = code_nested + children_code_nested

        return code_nested

    def export_widget_for_app_template(self, widget, first_node=False):
        if first_node:
            self.code_declared_classes = {}
            self.pending_listener_registration = list()
            # a list containing widgets that have been parsed and that are considered valid listeners
            self.known_project_children = [self, ]
            # a list containing dicts {listener, emitter, register_function, listener_function}
            self.pending_signals_to_connect = list()
            self.path_to_this_widget = []
            self.prepare_path_to_this_widget(self.children['root'])
        self.known_project_children.append(widget)

        widget.path_to_this_widget.append(widget.variable_name)

        code_nested = ''  # the code strings to return

        if not hasattr(widget, 'attributes'):
            return ''  # no nested code

        widgetVarName = widget.variable_name
        classname = 'CLASS' + \
            widgetVarName if widget.attr_editor_newclass else widget.__class__.__name__

        if not first_node:
            code_nested = prototypes.proto_widget_allocation % {
                'varname': widgetVarName, 'classname': classname}

        for x, y in inspect.getmembers(widget.__class__):
            if type(y) == property and not getattr(widget, x) is None:
                if hasattr(y.fget, "editor_attributes"): #if this property is visible for the editor
                    _value = getattr(widget, x)
                    if type(_value) == str:
                        _value = '"%s"' % _value
                    code_nested += prototypes.proto_property_setup % {
                        'varname': 'self' if first_node else widgetVarName, 'property': x, 'value': _value}
        else:
            #eventually, properties have to be placed in the container constructor
            pass

        # for all the methods of this widget
        for (setOnEventListenerFuncname, setOnEventListenerFunc) in inspect.getmembers(widget):
            # if the member is decorated by decorate_set_on_listener
            if issubclass(type(setOnEventListenerFunc), gui.ClassEventConnector):
                # if there is a callback
                if not setOnEventListenerFunc.callback is None and hasattr(setOnEventListenerFunc.event_method_bound, '_event_info'):
                    listenerFunction = setOnEventListenerFunc.callback
                    if issubclass(type(listenerFunction), gui.ClassEventConnector):
                        listenerFunction = listenerFunction.event_method_bound

                    listenerPrototype = setOnEventListenerFunc.event_method_bound._event_info[
                        'prototype']
                    listener = listenerFunction.__self__

                    # setOnEventListenerFunc._event_info['name'] + "_" + widget.identifier
                    listenerFunctionName = listenerFunction.__name__

                    listenerClassFunction = prototypes.proto_code_function % {'funcname': listenerFunctionName,
                                                                              'parameters': listenerPrototype}

                    # override, if already implemented, we use this code, unless it is a fakeListenerFunction
                    if hasattr(listener, listenerFunctionName) and listenerFunction.__code__ != editor_widgets.fakeListenerFunc.__code__:
                        listenerClassFunction = inspect.getsource(
                            listenerFunction)

                    # if the listener is already in the dictionary, and so it already used by another event
                    # skip
                    skip = False
                    for pending in self.pending_listener_registration:
                        if pending['eventlistener'] == listener:
                            if pending['listenerfuncname'] == listenerFunctionName:
                                skip = True
                                break
                    #if skip:
                    #    continue

                    self.pending_listener_registration.append({'done': False,
                                                               'eventsource': widget,
                                                               'eventlistener': listener,
                                                               'setoneventfuncname': setOnEventListenerFuncname,
                                                               'listenerfuncname': listenerFunctionName,
                                                               'listenerClassFunction': listenerClassFunction,
                                                               'skip_function_definition':skip})

        if widget.attr_editor_newclass or first_node:
            widgetVarName = 'self'

        children_code_nested = ''
        for child_key in widget.children.keys():
            child = widget.children[child_key]
            if type(child) == str:
                #children_code_nested += prototypes.proto_layout_append%{'parentname':widgetVarName,'varname':"'%s'"%child}
                continue
            if not issubclass(child.__class__, gui.Widget):
                continue
            if child.variable_name is None:
                continue
            child.path_to_this_widget = widget.path_to_this_widget[:]
            children_code_nested += self.repr_widget_for_editor(child)
            children_code_nested += prototypes.proto_layout_append % {
                'parentname': widgetVarName, 'varname': "%s,'%s'" % (child.variable_name, child.variable_name)}

        events_registration = self.check_pending_listeners(
            widget, widgetVarName)


        # and not (classname in self.code_declared_classes.keys()):
        if widget.attr_editor_newclass or first_node:
            if not widget.identifier in self.code_declared_classes:
                self.code_declared_classes[widget.identifier] = ''
            if first_node:
                if len(events_registration) < 1:
                    events_registration = 'pass'
                self.code_declared_classes[widget.identifier] = prototypes.proto_export_app_template % {'classname': classname, 'superclassname': widget.__class__.__name__,
                                                                                           'nested_code': code_nested + children_code_nested, 'events_registration': events_registration} + self.code_declared_classes[widget.identifier]
                code_nested = ''                                                                                           
            else:
                children_code_nested += events_registration
                self.code_declared_classes[widget.identifier] = prototypes.proto_code_class % {'classname': classname, 'superclassname': widget.__class__.__name__,
                                                                                           'nested_code': children_code_nested} + self.code_declared_classes[widget.identifier]
        else:
            code_nested = code_nested + children_code_nested

        return code_nested

    def prepare_path_to_this_widget(self, node):
        # here gets initiated to null list the path_to_this_widget chain
        node.path_to_this_widget = []
        for child in node.children.values():
            if type(child) == str:
                continue
            if not issubclass(child.__class__, gui.Widget):
                continue
            if child.variable_name is None:
                continue
            self.prepare_path_to_this_widget(child)

    def save(self, save_path_filename, configuration):
        compiled_code = ''
        code_classes = ''

        ret = self.repr_widget_for_editor(self.children['root'], True)
        # + self.code_listener_registration[str(id(self))]
        code_nested = ret + self.check_pending_listeners(self, 'self', True)

        main_code_class = prototypes.proto_code_main_class % {'classname': configuration.configDict[configuration.KEY_PRJ_NAME],
                                                              'config_resourcepath': configuration.configDict[configuration.KEY_RESOURCEPATH],
                                                              'code_nested': code_nested,
                                                              'mainwidgetname': self.children['root'].variable_name}

        if self.identifier in self.code_declared_classes.keys():
            main_code_class += self.code_declared_classes[self.identifier]
            del self.code_declared_classes[self.identifier]

        for key in self.code_declared_classes.keys():
            code_class = self.code_declared_classes[key]
            code_listener_setting = ''
            code_classes += code_class

        modules_to_import = []
        for w in self.known_project_children:
            if not w.__module__ in modules_to_import and w.__module__ != "__main__" and w.__module__ != "project":
                modules_to_import.append(w.__module__)

        code_classes += main_code_class
        compiled_code = prototypes.proto_code_program % {'imports': '\n'.join(["from " + modulename + " import *" for modulename in modules_to_import]),
                                                         'code_classes': code_classes,
                                                         'classname': configuration.configDict[configuration.KEY_PRJ_NAME],
                                                         'configuration': configuration.configDict
                                                         }

        print(compiled_code)

        if save_path_filename != None:
            f = open(save_path_filename, "w")
            f.write(compiled_code)
            f.close()


class Editor(App):
    EVENT_ONDROPPPED = "on_dropped"

    selectedWidget = None
    additional_widgets_loaded = False

    def __init__(self, *args):
        editor_res_path = os.path.join(os.path.dirname(__file__), 'res')
        super(Editor, self).__init__(
            *args, static_file_path={'editor_resources': editor_res_path})

    def idle(self):
        for drag_helper in self.drag_helpers:
            drag_helper.update_position()

    def main(self):
        import time
        t= time.time()
        #custom css
        my_css_head = """
            <link href='/editor_resources:style.css' rel='stylesheet' />
            """
        self.page.children['head'].add_child('mycss', my_css_head)

        self.mainContainer = gui.GridBox(width='100%', height='100%') 
        self.mainContainer.set_from_asciiart("""
            | menubar  | menubar                     | instances  |
            | toolbox  | toolbar                     | instances  |
            | toolbox  | project                     | instances  |
            | toolbox  | project                     | instances  |
            | toolbox  | project                     | instances  |
            | toolbox  | project                     | instances  |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | project                     | properties |
            | toolbox  | signals                     | properties |
            | toolbox  | signals                     | properties |
            | toolbox  | signals                     | properties |
            | toolbox  | signals                     | properties |
            | toolbox  | signals                     | properties |
            """, 0, 0)

        menubar = gui.MenuBar(width='100%', height='100%')
        menu = gui.Menu(width='100%', height='100%')
        menu.style['z-index'] = '1'
        m1 = gui.MenuItem('File', width=150, height='100%')
        m10 = gui.MenuItem('New', width=150, height=30)
        m11 = gui.MenuItem('Open', width=150, height=30)
        m12 = gui.MenuItem('Save Your App', width=150, height=30)
        #m12.style['visibility'] = 'hidden'
        m121 = gui.MenuItem('Save', width=100, height=30)
        m122 = gui.MenuItem('Save as', width=100, height=30)
        m123 = gui.MenuItem('Export widget as', width=200, height=30)
        m124 = gui.MenuItem('Export widget for App Template', width=200, height=30)
        m1.append([m10, m11, m12])
        m12.append([m121, m122, m123, m124])

        m2 = gui.MenuItem('Edit', width=100, height='100%')
        m21 = gui.MenuItem('Cut', width=100, height=30)
        m22 = gui.MenuItem('Paste', width=100, height=30)
        m2.append([m21, m22])

        m3 = gui.MenuItem('Project Config', width=200, height='100%')

        m4 = gui.MenuItem('Became a Sponsor', width=200,
                          height='100%', style={'font-weight': 'bold'})

        menu.append([m1, m2, m3, m4])

        menubar.append(menu)

        self.toolbar = editor_widgets.ToolBar(
            width='100%', height='100%', margin='0px 0px')
        self.toolbar.style['border-bottom'] = '1px solid rgba(0,0,0,.12)'
        self.toolbar.add_command(
            '/editor_resources:delete.png', self.toolbar_delete_clicked, 'Delete Widget')
        self.toolbar.add_command(
            '/editor_resources:cut.png', self.menu_cut_selection_clicked, 'Cut Widget')
        self.toolbar.add_command(
            '/editor_resources:paste.png', self.menu_paste_selection_clicked, 'Paste Widget')

        lbl = gui.Label("Snap grid", width=100)
        self.spin_grid_size = gui.SpinBox('15', '1', '100', width=50, height="100%")
        self.spin_grid_size.onchange.do(self.on_snap_grid_size_change)

        grid_size = gui.HBox(children=[lbl, self.spin_grid_size], style={
                             'outline': '1px solid gray', 'margin': '2px', 'margin-left': '10px'})
        self.toolbar.append(grid_size)

        self.fileOpenDialog = editor_widgets.EditorFileSelectionDialog(
            'Open Project', 'Select the project file.<br>It have to be a python program created with this editor.', False, '.', True, False, self)
        self.fileOpenDialog.confirm_value.do(self.on_open_dialog_confirm)

        self.fileSaveAsDialog = editor_widgets.EditorFileSaveDialog(
            'Project Save', 'Select the project folder and type a filename', False, '.', False, True, self)
        self.fileSaveAsDialog.add_fileinput_field('untitled.py')
        self.fileSaveAsDialog.confirm_value.do(self.menu_save_clicked)

        m10.onclick.do(self.menu_new_clicked)
        m11.onclick.do(self.fileOpenDialog.show)
        m121.onclick.do(self.menu_save_clicked)
        m122.onclick.do(self.fileSaveAsDialog.show)
        m123.onclick.do(self.menu_save_widget_clicked)
        m124.onclick.do(self.menu_export_widget_clicked)
        m21.onclick.do(self.menu_cut_selection_clicked)
        m22.onclick.do(self.menu_paste_selection_clicked)

        m3.onclick.do(self.menu_project_config_clicked)
        m4.onclick.do(self.menu_became_a_sponsor)

        # here are contained the widgets
        self.widgetsCollection = editor_widgets.WidgetCollection(
            self, width='100%', height='100%')

        self.projectConfiguration = editor_widgets.ProjectConfigurationDialog(
            'Project Configuration', 'Write here the configuration for your project.')

        self.signalConnectionManager = editor_widgets.SignalConnectionManager(
            width='100%', height='100%', style={'order': '1'})

        self.mainContainer.append( 
            {'toolbox': self.widgetsCollection, 'signals': self.signalConnectionManager, 'menubar':menubar})

        self.mainContainer.append(self.toolbar,'toolbar')

        self.instancesWidget = editor_widgets.InstancesWidget(width='100%')
        self.instancesWidget.treeView.on_tree_item_selected.do(
            self.on_instances_widget_selection)

        self.attributeEditor = editor_widgets.EditorAttributes(
            self, width='100%', height="100%")
        self.attributeEditor.style['overflow'] = 'auto'

        self.mainContainer.append(
            {'instances': self.instancesWidget, 'properties': self.attributeEditor})

        self.drag_helpers = [ResizeHelper(self, width=16, height=16),
                             DragHelper(self, width=15, height=15),
                             SvgDraggablePoint(
                                 self, 'cx', 'cy', [gui.SvgCircle]),
                             SvgDraggableCircleResizeRadius(
                                 self, [gui.SvgCircle]),
                             SvgDraggablePoint(
                                 self, 'x1', 'y1', [gui.SvgLine]),
                             SvgDraggablePoint(
                                 self, 'x2', 'y2', [gui.SvgLine]),
                             SvgDraggablePoint(
                                 self, 'x', 'y', [gui.SvgRectangle, gui.SvgText, gui.SvgImage, gui.SvgSubcontainer]),
                             SvgDraggableRectangleResizePoint(self, [gui.SvgRectangle, gui.SvgImage, gui.SvgSubcontainer])]
        for drag_helper in self.drag_helpers:
            drag_helper.stop_drag.do(self.on_drag_resize_end)

        self.menu_new_clicked(None)
        self.on_snap_grid_size_change(
            self.spin_grid_size, self.spin_grid_size.get_value())

        self.projectPathFilename = ''
        self.editCuttedWidget = None  # cut operation, contains the cutted tag
        print(">>>>>>>>>>>>startup time:", time.time()-t)
        # returning the root widget
        return self.mainContainer

    def on_snap_grid_size_change(self, emitter, value):
        value = float(value)
        for drag_helper in self.drag_helpers:
            drag_helper.set_snap_grid_size(value)

    def on_drag_resize_end(self, emitter):
        self.selectedWidget.__dict__[
            'properties'].set_widget(self.selectedWidget)

    def configure_widget_for_editing(self, widget):
        """ A widget have to be added to the editor, it is configured here in order to be conformant 
            to the editor
        """
        if not issubclass(widget.__class__, gui.Widget):
            return
        if widget.variable_name is None:
            return

        # for all the methods of this widget

        for (setOnEventListenerFuncname, setOnEventListenerFunc) in inspect.getmembers(widget):
            # if the member is decorated by decorate_set_on_listener
            if hasattr(setOnEventListenerFunc, '_event_info'):
                setOnEventListenerFunc.__class__ = editor_widgets.ClassEventConnectorEditor
                if setOnEventListenerFunc.callback:
                    getattr(widget, setOnEventListenerFuncname).do(
                        setOnEventListenerFunc.callback)

        # we force a connection, also if back_callback is None, to get the widget selectable
        back_callback = widget.onclick.callback
        widget.onclick.do(back_callback, js_stop_propagation=True)
        widget.onclick.editor_listener_callback = self.on_widget_selection

        # setup of the on_dropped function of the widget in order to manage the dragNdrop
        widget.__class__.on_dropped = on_dropped

        # drag properties
        #widget.style['resize'] = 'both'
        #widget.style['overflow'] = 'auto'
        widget.attributes['draggable'] = 'true'

        widget.attributes['tabindex'] = str(self.tabindex)
        # if not 'position' in widget.style.keys():
        #    widget.style['position'] = 'absolute'
        # if not 'left' in widget.style.keys():
        #    widget.style['left'] = '1px'
        # if not 'top' in widget.style.keys():
        #    widget.style['top'] = '1px'
        self.tabindex += 1

    def add_widget_to_editor(self, widget, parent=None, root_tree_node=True):
        if parent == None:
            parent = self.selectedWidget

        self.configure_widget_for_editing(widget)
        #widget.identifier = widget.attributes.get('editor_varname', widget.identifier)
        key = widget.identifier
        if hasattr(widget, 'variable_name'):
            key = widget.variable_name
        key = "root" if parent == self.project else key
        if root_tree_node:
            parent.append(widget, key)
            if self.selectedWidget == self.project:
                self.on_widget_selection(widget)
        #dcopy = widget.children.copy()
        for child in widget.children.values():
            if type(child) == str:
                continue
            self.add_widget_to_editor(child, widget, False)
        self.instancesWidget.update(self.project, self.selectedWidget)
        # if root_tree_node:
        #    self.on_widget_selection(widget)

    def on_instances_widget_selection(self, instancesWidgetItem, selectedWidget):
        self.on_widget_selection(selectedWidget)

    def on_widget_selection(self, widget):
        import time
        t = time.time()

        if issubclass(widget.__class__, gui.Container) or widget == None:
            self.mainContainer.append(
                self.widgetsCollection, 'toolbox')
        else:
            self.mainContainer.append(gui.Label("Cannot append widgets to %s class. It is not a container. Select a container" %
                                                   widget.__class__.__name__), 'toolbox')

        if self.selectedWidget == widget or widget == self.project:
            self.selectedWidget = widget
            return
        self.remove_box_shadow_selected_widget()
        self.selectedWidget = widget

        self.selectedWidget.style['box-shadow'] = '0 0 10px rgb(33,150,243)'
        # if self.selectedWidget.__dict__.get('signal_manager',None) is None:
        #    self.selectedWidget.__dict__['signal_manager'] = editor_widgets.SignalConnectionManager(width='100%', height='50%', style={'order':'1'})
        self.signalConnectionManager.update(self.selectedWidget, self.project)
        #self.subContainerLeft.append(self.selectedWidget.__dict__['signal_manager'], 'signal_manager')
        if self.selectedWidget.__dict__.get('properties', None) is None:
            self.selectedWidget.__dict__['properties'] = editor_widgets.EditorAttributes(
                self, width='100%', height='100%', style={'overflow': 'auto'})
            self.selectedWidget.__dict__[
                'properties'].set_widget(self.selectedWidget)
        self.mainContainer.append(self.selectedWidget.__dict__[
                                      'properties'], 'properties')

        parent = self.selectedWidget.get_parent()
        for drag_helper in self.drag_helpers:
            drag_helper.setup(widget, parent)
        # self.instancesWidget.select(self.selectedWidget)
        self.instancesWidget.update(self.project, self.selectedWidget)
        print("selected widget: " + widget.variable_name)
        print("selected widget class: " + widget.__class__.__name__)
        print("is widget Container: " +
              str(issubclass(self.selectedWidget.__class__, gui.Container)))

        print(time.time()-t)

    def menu_new_clicked(self, widget):
        print('new project')
        # self.project.new()
        self.project = Project(width='100%', height='100%')
        self.project.style['min-height'] = '400px'
        self.project.attributes['ondragover'] = "event.preventDefault();"
        self.project.attributes['ondrop'] = """event.preventDefault();
                var data = JSON.parse(event.dataTransfer.getData('application/json'));
                var params={};
                if( data[0] == 'add'){
                    params['left']=event.clientX-event.currentTarget.getBoundingClientRect().left;
                    params['top']=event.clientY-event.currentTarget.getBoundingClientRect().top;
                }
                remi.sendCallbackParam(data[1],'%(evt)s',params);
                
                return false;""" % {'evt': self.EVENT_ONDROPPPED}
        self.project.onkeydown.do(self.onkeydown)
        self.mainContainer.append(self.project, 'project')
        self.project.style['position'] = 'relative'
        self.tabindex = 0  # incremental number to allow widgets selection
        self.selectedWidget = None
        self.on_widget_selection(self.project)
        for drag_helper in self.drag_helpers:
            drag_helper.setup(None, None)
        if 'root' in self.project.children.keys():
            self.project.remove_child(self.project.children['root'])

    def on_open_dialog_confirm(self, widget, filelist):
        if len(filelist):
            self.menu_new_clicked(None)
            try:
                widgetTree = self.project.load(
                    filelist[0], self.projectConfiguration)
                if widgetTree != None:
                    self.add_widget_to_editor(widgetTree)
                self.projectPathFilename = filelist[0]
            except Exception:
                self.show_error_dialog("ERROR: Unable to load the project",
                                       "There were an error during project load: %s" % traceback.format_exc())

    def menu_save_clicked(self, widget, path=""):
        # the dragHelper have to be removed
        for drag_helper in self.drag_helpers:
            drag_helper.setup(None, None)
        if len(path):
            self.projectPathFilename = path + '/' + \
                self.fileSaveAsDialog.get_fileinput_value()
        else:
            if not len(self.projectPathFilename):
                self.fileSaveAsDialog.confirm_value.do(self.menu_save_clicked)
                self.fileSaveAsDialog.show()
                return

        self.remove_box_shadow_selected_widget()
        self.project.save(self.projectPathFilename, self.projectConfiguration)

    def menu_save_widget_clicked(self, widget, path=""):
        """ This method allows to export the selected widget
        """
        if len(path):
            self.projectPathFilename = path + '/' + \
                self.fileSaveAsDialog.get_fileinput_value()
        else:
            self.fileSaveAsDialog.confirm_value.do(
                self.menu_save_widget_clicked)
            self.fileSaveAsDialog.show()
            return

        code = ""
        code = code + \
            self.project.repr_widget_for_editor(self.selectedWidget, True)
        for key in self.project.code_declared_classes.keys():
            code_class = self.project.code_declared_classes[key]
            code = code + code_class
        with open(self.projectPathFilename, "w") as f:
            f.write(code)

    def menu_export_widget_clicked(self, widget, path=""):
        """ This method allows to export the selected widget for app_template 
            https://bitbucket.org/cheak/app-template-for-remi/src/master/views/from_remi_editor2.py
        """
        if len(path):
            self.projectPathFilename = path + '/' + \
                self.fileSaveAsDialog.get_fileinput_value()
        else:
            self.fileSaveAsDialog.confirm_value.do(
                self.menu_export_widget_clicked)
            self.fileSaveAsDialog.show()
            return

        code = ""
        code = code + \
            self.project.export_widget_for_app_template(self.selectedWidget, True)
        for key in self.project.code_declared_classes.keys():
            code = code + self.project.code_declared_classes[key]
        with open(self.projectPathFilename, "w") as f:
            f.write(code)

    def remove_box_shadow_selected_widget(self):
        if 'box-shadow' in self.selectedWidget.style.keys():
            del self.selectedWidget.style['box-shadow']

    def menu_cut_selection_clicked(self, widget):
        if self.selectedWidget == self.project:
            return
        for drag_helper in self.drag_helpers:
            drag_helper.setup(None, None)
        parent = self.selectedWidget.get_parent()
        self.editCuttedWidget = self.selectedWidget
        parent.remove_child(self.selectedWidget)
        print("tag cutted:" + self.editCuttedWidget.variable_name)
        self.on_widget_selection(parent)

    def menu_paste_selection_clicked(self, widget):
        if self.editCuttedWidget != None:
            key = "root" if self.selectedWidget == self.project else self.editCuttedWidget.identifier
            self.selectedWidget.append(self.editCuttedWidget, key)
            self.editCuttedWidget = None
            self.instancesWidget.update(self.project, self.selectedWidget)

    def menu_became_a_sponsor(self, widget):
        dialog = gui.GenericDialog(
            "Became a Sponsor", "This editor is made for you with passion. \nIt would be fantastic if you give a contribution, also a little one. ;-)", width="700px")

        sponsor_card = gui.Widget(_type='iframe', width=610, height=230) 
        #<iframe src="https://github.com/sponsors/dddomodossola/card" title="Sponsor dddomodossola" height="225" width="600" style="border: 0;"></iframe>
        sponsor_card.attributes['src'] = "https://github.com/sponsors/dddomodossola/card"
        sponsor_card.attributes['title'] = "Sponsor dddomodossola"
        sponsor_card.attributes['height'] = "225"
        sponsor_card.attributes['width'] = "600"
        sponsor_card.style['border'] = "0"

        dialog.add_field("sponsor card", sponsor_card)
        dialog.children["message"].style['white-space'] = 'pre'
        dialog.cancel.style['display'] = 'none'
        dialog.conf.set_text("Back")
        dialog.show(self)

    def menu_project_config_clicked(self, widget):
        self.projectConfiguration.show(self)

    def toolbar_delete_clicked(self, widget):
        if self.selectedWidget == self.project:
            return
        for drag_helper in self.drag_helpers:
            drag_helper.setup(None, None)
        parent = self.selectedWidget.get_parent()
        parent.remove_child(self.selectedWidget)
        todel = self.selectedWidget
        self.on_widget_selection(parent)
        del todel
        print("tag deleted")

    def move_widget(self, css_key, value):
        # css_key can be 'top' or 'left'
        # value (int): positive or negative value
        if issubclass(self.selectedWidget.__class__, gui.Widget) and css_key in self.selectedWidget.style:
            self.selectedWidget.style[css_key] = gui.to_pix(gui.from_pix(self.selectedWidget.style[css_key]) + value)

    def onkeydown(self, emitter, key, keycode, ctrl, shift, alt):
        arrow_left = '37'
        arrow_right = '39'
        arrow_up = '38'
        arrow_down = '40'
        key_canc = '46'
        ctrl = ctrl.lower() == 'true'
        shift = shift.lower() == 'true'
        if str(keycode) == key_canc:  # 46 the delete keycode
            self.toolbar_delete_clicked(None)
            return

        if keycode in (arrow_left, arrow_right, arrow_up, arrow_down):
            value = int(self.spin_grid_size.get_value())
            value = value if str(keycode) in (arrow_down, arrow_right) else -value
            key = 'left' if str(keycode) in (arrow_left, arrow_right) else 'top'
            if ctrl:
                key = {'left':'width', 'top':'height'}[key]

            self.move_widget(key, value)
        
        self.on_drag_resize_end(self)

        print("Key pressed: " + str(keycode) + "  ctrl: " + str(ctrl) + "  shift: " + str(shift))

    def show_error_dialog(self, title, message):
        error_dialog = gui.GenericDialog(title, message)
        error_dialog.children["title"].style['background-color'] = 'red'
        error_dialog.children["message"].style['white-space'] = 'pre'
        error_dialog.cancel.style['display'] = 'none'
        error_dialog.show(self)

    def onload(self, emitter):
        if not self.additional_widgets_loaded:
            print("loading additional widgets")
            self.additional_widgets_loaded = True
            self.widgetsCollection.load_additional_widgets()


def on_dropped(self, left, top):
    if len(left) < 1:
        left = '0px'
    if len(top) < 1:
        top = '0px'
    self.style['left'] = left
    self.style['top'] = top


if __name__ == "__main__":
    start(Editor, debug=False, address='0.0.0.0', port=8082,
          update_interval=0.05, multiple_instance=True)



# -*- coding: utf-8 -*-

import remi
import remi.gui as gui
from remi.gui import *
import snap7
import threading
from threading import Timer
import traceback

# https://python-snap7.readthedocs.io/en/latest/util.html
# https://github.com/gijzelaerr/python-snap7/blob/master/example/boolean.py


style_inheritance_dict = {'opacity':'inherit', 'overflow':'inherit', 'background-color':'inherit', 'background-image':'inherit', 'background-position':'inherit', 'background-repeat':'inherit', 'border-color':'inherit', 'border-width':'inherit', 'border-style':'inherit', 'border-radius':'inherit', 'color':'inherit', 'font-family':'inherit', 'font-size':'inherit', 'font-style':'inherit', 'font-weight':'inherit', 'white-space':'inherit', 'letter-spacing':'inherit'}
style_inheritance_text_dict = {'opacity':'inherit', 'overflow':'inherit', 'color':'inherit', 'font-family':'inherit', 'font-size':'inherit', 'font-style':'inherit', 'font-weight':'inherit', 'white-space':'inherit', 'letter-spacing':'inherit'}


class PLCSiemens(Image):
    """ This is a snap7 interface that allows to communicate with Siemens PLC.
        It handles the connection and reconnection.
        Widgets' event "link_to" have to be connected to this widget on the "on_link_to" method.
        i.e.:
            widget.link_to.do(plc.on_link_to)
    """
    icon = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABcAAAAvCAYAAAAIA1FgAAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAADmwAAA5sBPN8HMQAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAANGSURBVFiF7ZbNbxtFGIefmR2vP+PUjkntxHGapmmpKpIoECggjpyQEBKqeoIjXDhw5cD/wZ0LEgiJEzck1EIrlUYBl7RJ1bQlbUpCHNux15+7MxycWnLzoX6sJZDyk1bzzsc+s/q9s7uv+PTzL40dCBCLRShsl/A8j1K5AoARnet5JAyoWDSC1hpjIDuS5sH6390Fi+9WyL49zungCPdam+TsFAMyjKObLNRXmbTTTIdyXKrdIiJsitohq4YoeBUWf19C3fvr4b47G2MAmLTTnLLTDMgQuUCKpeYDzgZHOR+eYt0tMmmn+cm5wRfDF/mm/AspK85qe4NFQBKuQ7gOyu2BFzYfAeChSVhRbKGomxbbXpU7rQ1iMoRGs+mWcY3mu52rvBaepKirnLYzHWsu3MAAmF/nMZfe6MK3NtZpvpOkOOYSt6N4xiNk2QQtm1KzihAgkcSDEZx2g4bXRglJW7sIIVGPXNRhSRldCTG68rinAA00SPbc1iABQGC3b+22AZT+/r1dH44dts9zSXF7wndoFy4++BEAs3wKbk75DD9zpxMVkp3M+gk3q7lOtN0Hz8237/sO7cLFiTUATGkQSnGf4Rd/6ERPvER+SPpKe0LKfPUxAKYR7AO87K/PPXD5ydcAmIVpzG8z/sJJlDtRuOErGECZq3OdaG2kZ0JaL55rZX5+a9+JZCr9wvC+HsX+nvP7q7do1mv+gwM2avzky76DH6u/tuw3eH5+FsepUW80GRtNc/nKdeZmzlGuVKlWHWZfOcv9tYfUG01i0QiBgOLaQv7p4JZloZRiKBkmffwl5mbOIaXg5IksrVabfwrb5JdWeHN+FtfTLOZv7vvke2yxLInahTtOjT/+XCYSCaGNYfXuGuWdCrVaHYBK1UFKweuvTiPE3qJSfPjRZ37/Ors6sCjKZTMMxgeoVB3GshkuX7nOYHyA1FCCUnmHWDRKq9VCStlTvD4VfDRzHBVQ5JdWyKSHMcbgeR5npibYKhS764QQzw53PY/EsUFGMsNsbG4xMZ4lFArSarURQhAOh2i32wgOLuAPhF9byKOUhet6PePLt++itUZK2S2znxkO7AEDaK172sP0//0qHsGP4Efw/zL8X7xWNIa0/NaYAAAAAElFTkSuQmCC"

    #OpencvImage inherits gui.Image and so it already inherits attr_src. 
    # I'm redefining it in order to avoid editor_attribute_decorator
    # and so preventing it to be shown in editor.
    @property
    def attr_src(self): return self.attributes.get('src', '')
    @attr_src.setter
    def attr_src(self, value): self.attributes['src'] = str(value)

    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The IP address as string', str, {})
    def ip_address(self): return self.__dict__.get('__ip_address', '127.0.0.1')
    @ip_address.setter
    def ip_address(self, v): 
        self.__dict__['__ip_address'] = v
        self.disconnect()
        self.connect()

    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The rack number as integer', int, {'possible_values': '', 'min': -1, 'max': 65535, 'default': 0, 'step': 1})
    def rack(self): return self.__dict__.get('__rack', -1)
    @rack.setter
    def rack(self, v): 
        self.__dict__['__rack'] = v
        self.disconnect()
        self.connect()

    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The slot number as integer', int, {'possible_values': '', 'min': -1, 'max': 65535, 'default': 0, 'step': 1})
    def slot(self): return self.__dict__.get('__slot', -1)
    @slot.setter
    def slot(self, v): 
        self.__dict__['__slot'] = v
        self.disconnect()
        self.connect()
        
    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The update interval in seconds as float', float, {'possible_values': '', 'min': 0.0, 'max': 655350.0, 'default': 1.0, 'step': 1})
    def update_interval(self): return self.__dict__.get('__update_interval', 1.0)
    @update_interval.setter
    def update_interval(self, v): 
        self.__dict__['__update_interval'] = v
        #self.disconnect()
        #self.connect()

    snap7_client = snap7.client.Client()    #the snap7.client.Client() instance that manages data exchange with PLC  
    connected = False
    linked_widgets = []                     #the widget list that are linked to the PLC
    update_lock = threading.RLock()
    app_instance = None

    def __init__(self, ip_address='', rack=0, slot=3, update_interval=1.0, *args, **kwargs):
        self.__ip_address = ''
        default_style = {'position':'absolute','left':'10px','top':'10px'}
        default_style.update(kwargs.get('style',{}))
        kwargs['style'] = default_style
        kwargs['width'] = kwargs['style'].get('width', kwargs.get('width','23px'))
        kwargs['height'] = kwargs['style'].get('height', kwargs.get('height','47px'))
        super(PLCSiemens, self).__init__(self.icon, *args, **kwargs)
        self.on_disconnected()
        self._set_params()
        self.__rack = rack
        self.__slot = slot
        if len(ip_address):
            self.ip_address = ip_address
        else:
            self.__ip_address = ip_address
        self.__update_interval = update_interval
        self.check_connection_state()

    def disconnect(self):
        try:
            self.snap7_client.disconnect()
        except Exception:
            print(traceback.format_exc())

    def connect(self):
        try:
            if self.rack<0 or self.slot<0:
                return
            print("connecting to ip:%s  rack:%s  slot:%s"%(self.ip_address, self.rack, self.slot))
            self.snap7_client.connect(self.ip_address, self.rack, self.slot)
            
        except Exception:
            print(traceback.format_exc())

    def _set_params(self):
        values = ()
        """(
                (snap7.snap7types.PingTimeout, 1000),
                (snap7.snap7types.SendTimeout, 500),
                (snap7.snap7types.RecvTimeout, 3500),
                (snap7.snap7types.SrcRef, 128),
                (snap7.snap7types.DstRef, 128),
                (snap7.snap7types.SrcTSap, 128),
                (snap7.snap7types.PDURequest, 470)
            )
        """
        for param, value in values:
            self.snap7_client.set_param(param, value)

    def get_cpu_info(self):
        """ i.e.
            ('ModuleTypeName', 'CPU 315-2 PN/DP'),
            ('SerialNumber', 'S C-CXXXXXXXXXXX'),
            ('ASName', 'SNAP7-SERVER'),
            ('Copyright', 'Original Siemens Equipment'),
            ('ModuleName', 'CPU 315-2 PN/DP')
        """
        fields = ('ModuleTypeName','SerialNumber','ASName','Copyright','ModuleName')
        cpuInfo = self.snap7_client.get_cpu_info()
        info_string = "IP address: %s\n"%self.ip_address
        for field in fields:
            info_string = info_string + "%s: %s\n"%(field,str(getattr(cpuInfo, field, '')))
        return info_string

    @decorate_event
    def on_connected(self):
        self.attributes['title'] =  "IP address: %s"%self.ip_address #self.get_cpu_info()
        return ()

    @decorate_event
    def on_disconnected(self):
        self.attributes['title'] = "%s : not connected"%self.ip_address
        return ()

    def __del__(self):
        self.update_interval = 0
        self.snap7_client.destroy()

    def search_app_instance(self, node):
        if issubclass(node.__class__, remi.server.App):
            return node
        if not hasattr(node, "get_parent"):
            return None
        return self.search_app_instance(node.get_parent()) 

    def check_connection_state(self):
        _con = self.snap7_client.get_connected()
        if _con != self.connected:
            if _con:
                self.on_connected()
            else:
                self.on_disconnected()
        self.connected = _con

        if self.app_instance==None:
            self.app_instance = self.search_app_instance(self)
            
        if self.app_instance:
            with self.update_lock:
                with self.app_instance.update_lock:
                    for w in self.linked_widgets:
                        if hasattr(w, "update"):
                            try:
                                w.update()
                            except Exception:
                                print(traceback.format_exc())

        if self.update_interval>0.0:
            Timer(self.update_interval, self.check_connection_state).start()

    def set_bool(self, db_index, byte_index, bit_index, value):
        reading = self.snap7_client.db_read(db_index, byte_index, 1)    # read 1 byte from db 31 staring from byte 120
        snap7.util.set_bool(reading, 0, bit_index, value)   # set a value of fifth bit
        self.snap7_client.db_write(db_index, byte_index, reading)    # write back the bytearray and now the boolean value is changed 

    def get_bool(self, db_index, byte_index, bit_index):
        bytes_to_read_write = 1
        mem = self.snap7_client.db_read(db_index, byte_index, bytes_to_read_write)
        return snap7.util.get_bool(mem, byte_index, bit_index)

    def get_byte(self, db_index, byte_index):
        bytes_to_read_write = 1
        mem = self.snap7_client.db_read(db_index, byte_index, bytes_to_read_write)
        return bytearray(mem)

    def set_int(self, db_index, byte_index, value):
        mem = self.snap7_client.db_read(db_index, byte_index, 2)
        snap7.util.set_int(mem, 0, value)
        self.snap7_client.db_write(db_index, byte_index, mem)

    def get_int(self, db_index, byte_index):
        mem = self.snap7_client.db_read(db_index, byte_index, 2) #an int from 2 bytes
        return snap7.util.get_int(mem, 0) #0 or byte_index?

    def set_real(self, db_index, byte_index, value):
        mem = self.snap7_client.db_read(db_index, byte_index, 4)
        snap7.util.set_real(mem, 0, value)
        self.snap7_client.db_write(db_index, byte_index, mem)

    def get_real(self, db_index, byte_index):
        mem = self.snap7_client.db_read(db_index, byte_index, 4) #a real from 4 bytes
        return snap7.util.get_real(mem, 0) #0 or byte_index?

    def set_string(self, db_index, byte_index, max_size, value):
        mem = self.snap7_client.db_read(db_index, byte_index, max_size)
        snap7.util.set_string(mem, 0, value, max_size)
        self.snap7_client.db_write(db_index, byte_index, mem)

    def get_string(self, db_index, byte_index, max_size):
        mem = self.snap7_client.db_read(db_index, byte_index, max_size)
        return snap7.util.get_string(mem, 0, max_size) #0 or byte_index?

    def on_link_to(self, widget):
        """ Each widget registers itself to the PLC class by linking its "link_to" event 
            toward the "on_link_to" listener
        """
        with self.update_lock:
            widget.set_plc_instance(self)
            #the widget is appended to a link for cyclical update purpose
            self.linked_widgets.append(widget)

    def remove_link_to(self, widget):
        with self.update_lock:
            self.linked_widgets.remove(widget)


# noinspection PyUnresolvedReferences
class SiemensWidget(object):
    plc_instance = None

    def _setup(self):
        #this must be called after the Widget super constructor
        self.link_to.do = self.do

    def do(self, callback, *userdata):
        if self.plc_instance and (callback is None):
            print("removing connection to the PLC")
            self.plc_instance.remove_link_to(self)
            self.plc_instance = None

        if hasattr(self.link_to.event_method_bound, '_js_code'):
            self.link_to.event_source_instance.attributes[self.link_to.event_name] = self.link_to.event_method_bound._js_code%{
                'emitter_identifier':self.link_to.event_source_instance.identifier, 'event_name':self.link_to.event_name}
        self.link_to.callback = callback
        self.link_to.userdata = userdata
        self.link_to.kwuserdata = {}
        #here the callback is called immediately to make it possible link to the plc
        if callback is not None: #protection against the callback replacements in the editor
            callback(self, *userdata)

    @decorate_set_on_listener("(self, emitter)")
    @decorate_event
    def link_to(self):
        return ()

    def set_plc_instance(self, plc_instance):
        self.plc_instance = plc_instance


class _Mixin_DB_property():
    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The DB number as integer', int, {'possible_values': '', 'min': -1, 'max': 65535, 'default': 0, 'step': 1})
    def db_index(self): return self.__dict__.get('__db_index', -1)
    @db_index.setter
    def db_index(self, v): self.__dict__['__db_index'] = v

class _Mixin_Byte_property():
    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The byte index in the DB as integer', int, {'possible_values': '', 'min': -1, 'max': 65535, 'default': 0, 'step': 1})
    def byte_index(self): return self.__dict__.get('__byte_index', -1)
    @byte_index.setter
    def byte_index(self, v): self.__dict__['__byte_index'] = v

class _Mixin_Bit_property():
    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The bit index in the byte as integer', int, {'possible_values': '', 'min': -1, 'max': 65535, 'default': 0, 'step': 1})
    def bit_index(self):  return self.__dict__.get('__bit_index', -1)
    @bit_index.setter
    def bit_index(self, v): self.__dict__['__bit_index'] = v


class SiemensButton(gui.Container, SiemensWidget, _Mixin_DB_property, _Mixin_Byte_property, _Mixin_Bit_property):
    """ A Button widget that sets the bit when clicked.
    """
    icon = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAC4AAAAuCAYAAABXuSs3AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAKMSURBVGiB7ZqxaxNRGMB/9+7uXZK2aaMVEUWEKNSh4CQIrhWnujiIQqGig7gXF/+B7tLNbp2KQyfBwcVdsdhWJa5WHZKKvZqXe3cO57VNmraJyd0ZuB88LnkvfO/H43sfB18Mz/OCIAhoN4DdZ9IYhrH7bDesSPJlRfHsrc+nmpGK6HGcGQp4clVwsywBMJRSgVKKqWXNpitS1juaS6M+L26ZSCnDE1dKsenaAHy/MUNjtNJ10JG1WYofHvTbtYnPWwKlFLZth+Ke5wGheKP0EXVireugemizz5rt8TyPIAgQe+KDQZO41jptn47RWofiAL7vp+3TMZGrtb9mAxTfP0bnf3QdMPf1Wt/kjiLytVoXRtZnEhHolf+7cB9BJp40mXjSDKz4gXLYHQ6vHtmUD8z7LC+4zAGz00M8PXv4q3Jl4xdTr7vfuUfx9pvP3xnm9v086893WFzZZjFamMzz7rpg7c02d1d72zOWVJn75oNjcDmO4H+JRXz+tKCyEaZKXPQlVcoTw3yZaJ776dpAox/h2xJLjocXUrI02eg5lw8jllRZXPGoYHBqPI7oIQNbx2MRn522KNc1S/9Qnzslpsvps7yws1e/Y6BH8TpTC/XOf766w5U+XdYsx5MmE0+aTDxpMvGkycSTRkTNoEEh8hXRl0Ehch3cEzcMA9M0GbdV2k7Hcj7/G9M098Qty+LhxSrnHDdtt0M5adW5d2ELy7LCU1dKBa7rUqvVqFaruK5LvV5Ptbvc2lV2HIdCoUCpVGJsbIxCoYAVLRSLRaSUKKVQStHaYkmDSFxKiZSSXC6H4zjhvOd5ge/7aK2bRtQkSruXL4TANM2mIYTA0FoHrWmR9h8QIlpTZv/nP6KyI2uh/zMtAAAAAElFTkSuQmCC"

    @property
    @gui.editor_attribute_decorator('WidgetSpecific','Specifies if the button is toggle or must reset the value on release', bool, {})
    def toggle(self): return self.__dict__.get('__toggle', None)
    @toggle.setter
    def toggle(self, v): self.__dict__['__toggle'] = v

    @property
    @editor_attribute_decorator("WidgetSpecific",'''Text content''', str, {})
    def text(self): return self.button.get_text()
    @text.setter
    def text(self, value): self.button.set_text(value)

    button = None   # The gui.Button widget instance
    led = None      # The led indicator Widget

    def __init__(self, button_label='siemens button', db_index=-1, byte_index=-1, bit_index=-1, toggle=False, *args, **kwargs):
        self.color_inactive = 'darkgray'
        self.color_active = 'rgb(0,255,0)'
        _style = style_inheritance_text_dict
        _style.update(style_inheritance_dict)
        self.button = gui.Button(button_label, width="100%", height="100%", style=_style)
        self.led = gui.Widget(width=15, height=5, style={'position':'absolute', 'left':'2px', 'top':'2px', 'background-color':self.color_inactive})
        self.led_status = False
        default_style = {'position':'absolute','left':'10px','top':'10px', 'background-color':'rgb(4, 90, 188)', 'color':'white'}
        default_style.update(kwargs.get('style',{}))
        kwargs['style'] = default_style
        kwargs['width'] = kwargs['style'].get('width', kwargs.get('width','100px'))
        kwargs['height'] = kwargs['style'].get('height', kwargs.get('height','100px'))
        super(SiemensButton, self).__init__(*args, **kwargs)
        SiemensWidget._setup(self)
        _style = {'position':'relative'}
        _style.update(style_inheritance_dict)
        self.append(gui.Container(children=[self.button, self.led], width="100%", height="100%", style=_style))
        self.toggle = toggle
        self.button.onmousedown.do(self.set_bit)
        self.button.onmouseup.do(self.reset_bit)

    def set_bit(self, emitter, *args, **kwargs):
        if self.db_index<0 or self.byte_index<0 or self.bit_index<0:
            return
        self.written = False
        value = 1
        if self.toggle:
            value = 0 if self.led_status else 1
        self.plc_instance.set_bool(self.db_index, self.byte_index, self.bit_index, value)

    def reset_bit(self, emitter, x, y, *args, **kwargs):
        if self.db_index<0 or self.byte_index<0 or self.bit_index<0:
            return
        if not self.toggle:
            self.plc_instance.set_bool(self.db_index, self.byte_index, self.bit_index, 0)

    def _set_value(self, value):
        #this function gets called when the camonitor notifies a change on the PV
        self.led_status = value
        self.led.style.update({'background-color':self.color_active if self.led_status else self.color_inactive})

    def update(self, *args):
        #this method gets called by the plc_instance
        if self.plc_instance==None:
            return
        if self.db_index<0 or self.byte_index<0 or self.bit_index<0:
            return
        value = self.plc_instance.get_bool(self.db_index, self.byte_index, self.bit_index)
        self._set_value(value)


class BitStatusWidget(HBox, SiemensWidget, _Mixin_DB_property, _Mixin_Byte_property, _Mixin_Bit_property):
    """A Status indicator widget.
        Connect the event link_to to a PLCSiemens widget on the "on_link_to" method.
        i.e.:
            widget.link_to.do(plc.on_link_to)
    """
    icon = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAEUAAAAdCAIAAABzMjbkAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA7DAcdvqGQAAAJrSURBVFhH3ZQxSBtRGMffcpDlpiylEG5xOYTQrUgGN5FMEjIJIhnawVGQuJxwk0OLHUKXgtCCZMgSh3aQFjoIVlBKaCDFQggR4pBMN914vvfu/9XLJTUqr5/gj/9wd3m53/e99+6JJ0jEC0nNB8DDBUnNB8DDBUnNB8DDBUnNB8DDBUnNB8DDBUnNB8DDBUnNB8CTpu27otzEzXTuMGQSks5KIDonwl8Tji2aqZ/+EQBPmkfup7kqnILwd8Rzce9+wovG+rxtyetMbrnWCnWhhC54YsTEkGZZuH47LjlxMzj2l3MZOcay5zeOBvJJ/JdkHbelLaTmnv30agvC3T4Poig49/KWUz1TlSQn/6zqWHkvHrHt/i08OWR6P99e2fZKvR9GYb9RzlrFgyFDP8ODopUtvjv5M5Le0WXvStY9Xmxw1buUP2pa3hw9n92PnojXhz/VK/ESSCeqmZ6H9CNVnfpGYS6bUbvJP1a7YrxYvW0cvd80d+5Hrsv3vdKLZ7Yl99t640LOSfyGZB235SH93Mx+0P20Ylurn9V1oli1gE7lsJtet5n93Kx2ODqVGzW/+5uhH7UpFt50pDfs18f6Wfyg1yr+wE5lz0H362beSvaDIfra9eRJEY5+vV+y437iifgiv5/4w+PpJwpb+yU6vQpb+hRSFb59KZ/p7sJWLT6lMrlyteKKpY/D9JBo+MNTN/ogqyzSYg2Otgp0vpX21cGolP+5H1ZIaj4AHi5Iaj4AHi5Iaj4AHi5Iaj4AHi5Iaj4AHi5Iaj4AHi5Iaj4AHi5Iaj4AHi5Iaj4AHi5Iaj4AHi5Iaj5PDiGuAaoeNYuSC/YWAAAAAElFTkSuQmCC"

    @property
    @gui.editor_attribute_decorator('WidgetSpecific','The label text', str, {})
    def text(self): return self.label.text
    @text.setter
    def text(self, v): self.label.text = v

    label = None        #a gui.Label widget that shows the fixed label text
    label_value = None  #a gui.Label widget that shows the value as '1' or '0'

    def __init__(self, text='bit status widget', db_index=-1, byte_index=-1, bit_index=-1, *args, **kwargs):
        """
        Args:
            text (str): The text that will be displayed on the label.
            kwargs: See Widget.__init__()
        """
        default_style = {'position':'absolute','left':'10px','top':'10px', 'align-items':'stretch', 'justify-content':'flex-start'}
        default_style.update(kwargs.get('style',{}))
        kwargs['style'] = default_style
        kwargs['width'] = kwargs['style'].get('width', kwargs.get('width','100px'))
        kwargs['height'] = kwargs['style'].get('height', kwargs.get('height','30px'))
        super(BitStatusWidget, self).__init__(*args, **kwargs)
        SiemensWidget._setup(self)
        _style = style_inheritance_text_dict
        _style.update(style_inheritance_dict)
        _style['border']  = '1px solid black'
        self.label = gui.Label(text, width="100%", height="100%", style=_style)
        _style.update({'background-color':'gray', 'text-align':'center'})
        self.label_value = gui.Label("0", width='30px', height="100%", style=_style)
        self.append([self.label, self.label_value])
        self.db_index = db_index
        self.byte_index = byte_index
        self.bit_index = bit_index

    def update(self, *args):
        if self.plc_instance==None:
            return
        if self.db_index<0 or self.byte_index<0 or self.bit_index<0:
            return
        value = self.plc_instance.get_bool(self.db_index, self.byte_index, self.bit_index)
        self.label_value.set_text( '1' if value else '0' )
        style={'border':'1px solid black', 'background-color':'gray'}
        if value:
            style={'border':'1px solid black', 'background-color':'yellow'}
        self.label_value.style.update(style)


class WordEditWidget(SpinBox, SiemensWidget, _Mixin_DB_property, _Mixin_Byte_property):
    """ A Word (16bit) value edit.
        Connect the event link_to to a PLCSiemens widget on the "on_link_to" method.
        i.e.:
            widget.link_to.do(plc.on_link_to)
    """
    icon = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAFkAAAAeCAIAAADINCUsAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA7DAcdvqGQAAAGjSURBVGhD7ZYxjoNADEWpt8wl0qbYHIT7pEqdg3CDvQJSirQUXIFy+x3P/MH2jBuKGSGtn1LAt4n4LyBl+HUy7oJxF4y7YNwF4y4Yd8F0dfFlgdkJ+Acu1mXZcFixLcuKQ3YxjQO4Pt7I3o8rsmEYpxjxWgYDMclJoAjRXpM22/F53S+352zo2Obn7XJ/fXCaXFBpKIj9441TD1mUHQHzKpHWIdpraN6W2LnUYYTVOyIaCqSYhNyjYz0NGCHaazBrS9HctFO5sFXULmRgXmOFaK/BrDl7f1tEQLigu99fbUWcyFgXpbNxIjvieitEew0tdyJaCFgiAuY7onrHR0AnWkW6JG/QNs2sEO01caEPR10U70MtIhBC+fhrNTizwm/UV2CjOYfeESBK5N+4QMkilBu6nqZGiPYabLSl6G/riC7kb7j3p7B6IohcViDs8GEdor0mLjTFbG6E+bmguwXp/qMKhbSFjoL9C/hhqEK016TNdhz8r9ULtNdg1o7D/8G7gPYazE6Au2C6ujg57oJxF4y7YNwF4y4Yd8EMP/PqH/rM6x+8Pwvc9mLA2gAAAABJRU5ErkJggg=="
    def __init__(self, db_index=-1, byte_index=-1, *args, **kwargs):
        """
        Args:
            kwargs: See Widget.__init__()
        """
        default_style = {'position':'absolute','left':'10px','top':'10px'}
        default_style.update(kwargs.get('style',{}))
        kwargs['style'] = default_style
        kwargs['width'] = kwargs['style'].get('width', kwargs.get('width','100px'))
        kwargs['height'] = kwargs['style'].get('height', kwargs.get('height','30px'))
        super(WordEditWidget, self).__init__(0, -32767, 32766, 1,*args, **kwargs)
        SiemensWidget._setup(self)
        self.db_index = db_index
        self.byte_index = byte_index
        self.onchange.do(self.on_changed)

    def on_changed(self, emitter, new_value):
        if self.plc_instance==None:
            return
        if self.db_index<0 or self.byte_index<0:
            return
        self.plc_instance.set_int( self.db_index, self.byte_index, int(self.get_value()))


class ByteViewWidget(Label, SiemensWidget, _Mixin_DB_property, _Mixin_Byte_property):
    """ A Byte (8bit) value view.
        Connect the event link_to to a PLCSiemens widget on the "on_link_to" method.
        i.e.:
            widget.link_to.do(plc.on_link_to)
    """
    #icon = default_icon("ByteViewWidget")

    @property
    def text(self): return self.get_text()
    @text.setter
    def text(self, value): self.set_text(value)

    def __init__(self, db_index=-1, byte_index=-1, *args, **kwargs):
        """
        Args:
            kwargs: See Widget.__init__()
        """
        default_style = {'position':'absolute','left':'10px','top':'10px'}
        default_style.update(kwargs.get('style',{}))
        kwargs['style'] = default_style
        kwargs['width'] = kwargs['style'].get('width', kwargs.get('width','100px'))
        kwargs['height'] = kwargs['style'].get('height', kwargs.get('height','30px'))
        super(ByteViewWidget, self).__init__(*args, **kwargs)
        SiemensWidget._setup(self)
        self.db_index = db_index
        self.byte_index = byte_index

    def update(self, *args):
        if self.plc_instance==None:
            return
        if self.db_index<0 or self.byte_index<0:
            return
        value = self.plc_instance.get_byte(self.db_index, self.byte_index)
        self.set_text( bin(value[0]) )
